import json
from dataclasses import asdict
from terka.domain import events
import logging


class BasePublisher:
    ...


class RedisPublisher(BasePublisher):

    def __init__(self,
                 client: "redis.Redis",
                 topic_prefix: str | None = None) -> None:
        self.client = client
        self.topic_prefix = topic_prefix

    def publish(self, topic: str, event: events.Event):
        if self.topic_prefix:
            topic = f"{self.topic_prefix}_{topic}"
        self.client.publish(topic, json.dumps(asdict(event)))


class LogPublisher(BasePublisher):

    def publish(self, topic: str, event: events.Event):
        logging.info("Published to topic '%s': %s", topic, asdict(event))





from enum import Enum, auto
from datetime import datetime


class EventType(Enum):
    STATUS = auto()
    DUE_DATE = auto()
    NAME = auto()
    PROJECT = auto()
    DESCRIPTION = auto()
    PRIORITY = auto()
    ASSIGNEE = auto()


class BaseEvent:

    def __init__(self,
                 event_type: str,
                 old_value: str,
                 new_value: str,
                 date: datetime = datetime.now()) -> None:
        if event_type.upper() not in [e.name for e in EventType]:
            raise ValueError(f"{event_type} is invalid EventType")
        self.type = event_type.upper()
        self.old_value = old_value
        self.new_value = new_value
        self.date = datetime.now()

    def __repr__(self) -> str:
        return (f"<{self.__class__.__name__}> [{self.type.name}] "
                f"{self.old_value} -> {self.new_value}")


class TaskEvent(BaseEvent):

    def __init__(self, task_id: int, event_type: str, old_value: str,
                 new_value: str, date: datetime = datetime.now()) -> None:
        self.task = task_id
        super().__init__(event_type=event_type,
                         old_value=old_value,
                         new_value=new_value,
                         date=date)


class ProjectEvent(BaseEvent):

    def __init__(self, project_id: int, event_type: str, old_value: str,
                 new_value: str, date: datetime = datetime.now()) -> None:
        self.project = project_id
        super().__init__(event_type=event_type,
                         old_value=old_value,
                         new_value=new_value,
                         date=date)



from typing import Any, Dict, List, Tuple, Optional, Union

from dataclasses import dataclass
from datetime import datetime, timedelta

import re
import sys, tempfile, os
from subprocess import run
import yaml

import logging
import rich
from rich.console import Console
from rich.table import Table

from terka.domain.task import Task
from terka.domain.project import Project
from terka.domain.user import User
from terka.domain.commentary import TaskCommentary, ProjectCommentary, EpicCommentary, StoryCommentary, SprintCommentary
from terka.domain.notes import BaseNote, TaskNote, ProjectNote, EpicNote, StoryNote, SprintNote
from terka.domain.tag import BaseTag, TaskTag, ProjectTag
from terka.domain.collaborators import TaskCollaborator, ProjectCollaborator
from terka.domain.event_history import TaskEvent, ProjectEvent
from terka.domain.sprint import Sprint, SprintTask
from terka.domain.time_tracker import TimeTrackerEntry
from terka.domain.epic import Epic, EpicTask
from terka.domain.story import Story, StoryTask
from terka.domain.external_connectors.asana import AsanaTask, AsanaProject
from terka.domain.workspace import Workspace

from terka.service_layer import services, printer
from terka.service_layer.ui import TerkaTask
from terka.adapters.repository import AbsRepository
from terka.utils import format_command, format_task_dict, convert_date, convert_status, update_task_dict

logger = logging.getLogger(__name__)


class TerkaException(Exception):
    ...


class TaskAddedToCompletedEntity(TerkaException):
    ...


class TaskAddedToEntity(TerkaException):
    ...


class EntityNotFound(TerkaException):
    ...


# @dataclass
# class CurrentEntry:
#     value: str
#     type: str
#     source: str = None


def new_task_template() -> str:
    return f"""
        # You are creating a task, enter below:
        ---
        status: BACKLOG
        name: 
        description: 
        project: 
        due_date: 
        priority: NORMAL
        sprints: 
        epics: 
        stories: 
        tags: 
        collaborators: 
        """

def new_composite_template(entity) -> str:
    return f"""
        # You are creating {entity.__name__}, enter below:
        ---
        name: 
        description: 
        project: 
        """


def new_composite_template(entity) -> str:
    return f"""
        # You are creating {entity.__name__}, enter below:
        ---
        name: 
        description: 
        project: 
        """


def new_sprint_template() -> str:
    today = datetime.now()
    next_monday = (today +
                   timedelta(days=(7 - today.weekday()))).strftime("%Y-%m-%d")
    next_sunday = (today +
                   timedelta(days=(13 - today.weekday()))).strftime("%Y-%m-%d")
    return f"""
        # You are creating a sprint, enter below:
        ---
        goal: 
        description: 
        start_date: {next_monday}
        end_date:  {next_sunday}
        """


def start_sprint_template(sprint: Sprint) -> str:
    ...


def edit_sprint_template(sprint: Sprint) -> str:
    return f"""
        # You are editing sprint {sprint.id}, enter below:
        ---
        goal: {sprint.goal}
        start_date: {sprint.start_date}
        end_date: {sprint.end_date}
        """


def edited_task_template(task: Task, project: str | None = None) -> str:
    return f"""
        # You are editing task {task.id}, enter below:
        ---
        status: {task.status.name}
        name: {task.name}
        description: {task.description if task.description else ""}
        project: {project or ""}
        sprints: {task.sprints[-1].sprint if task.sprints else ""}
        epics: {task.epics[-1].epic if task.epics else ""}
        stories: {task.stories[-1].story if task.stories else ""}
        tags: {task.tags.pop() if task.tags else ""}
        collaborators: {task.collaborators.pop() if task.collaborators else ""}
        time_spent: 0 (task total_time_spent {task.total_time_spent})
        comment: 
        """


def completed_task_template(task: Task, project: str | None = None) -> str:
    return f"""
        # You are closing task {task.id}, enter below:
        ---
        status: DONE 
        time_spent: 0 (task total_time_spent {task.total_time_spent})
        comment: 
        name: {task.name}
        project: {project or ""}
        description: {task.description if task.description else ""}
        sprints: {task.sprints[-1].sprint if task.sprints else ""}
        epics: {task.epics[-1].epic if task.epics else ""}
        stories: {task.stories[-1].story if task.stories else ""}
        tags: {task.tags.pop() if task.tags else ""}
        collaborators: {task.collaborators.pop() if task.collaborators else ""}
        """


def generate_message_template(task: Optional[Task] = None,
                              repo=None,
                              kwargs: Optional[Dict[str, str]] = None,
                              entity=None) -> str:
    if task:
        if isinstance(task, (Task, Story, Epic)):
            project = task.project
        elif isinstance(task, Project):
            project = task.name
    if isinstance(task, Sprint):
        if not kwargs:
            message_template = new_sprint_template()
        else:
            message_template = edit_sprint_template(task)
    else:
        if not kwargs:
            if entity == Task:
                message_template = new_task_template()
            if entity == Sprint:
                message_template = new_sprint_template()
            if entity in (Epic, Story):
                message_template = new_composite_template(entity)
        elif kwargs and kwargs.get("status"):
            message_template = completed_task_template(task, project)
        else:
            message_template = edited_task_template(task, project)
    return re.sub("\n +", "\n", message_template.lstrip())


class BaseHandler:

    def __init__(self, successor) -> None:
        self._successor = successor

    def handle(self, entity):
        if self._successor:
            return self._successor.handle(entity)
        return None


class TaskHandler(BaseHandler):

    def handle(self, entity):
        if "tasks".startswith(entity):
            logger.debug("Handling task")
            return Task, "tasks"
        return super().handle(entity)


class ProjectHandler(BaseHandler):

    def handle(self, entity):
        if "projects".startswith(entity):
            logger.debug("Handling project")
            return Project, "projects"
        return super().handle(entity)


class UserHandler(BaseHandler):

    def handle(self, entity):
        if "users".startswith(entity):
            logger.debug("Handling user")
            return User, "users"
        return super().handle(entity)


class TagHandler(BaseHandler):

    def handle(self, entity):
        if "tags".startswith(entity):
            logger.debug("Handling tags")
            return BaseTag, "tags"
        return super().handle(entity)


class SprintHandler(BaseHandler):

    def handle(self, entity):
        if "sprints".startswith(entity):
            logger.debug("Handling sprints")
            return Sprint, "sprints"
        return super().handle(entity)


class SprintTaskHandler(BaseHandler):

    def handle(self, entity):
        if entity == "sprint_tasks":
            logger.debug("Handling sprint task")
            return SprintTask, "sprint_tasks"
        return super().handle(entity)


class TimeTrackerHandler(BaseHandler):

    def handle(self, entity):
        if entity == "time_entry":
            logger.debug("Handling time entry")
            return TimeTrackerEntry, "time_entries"
        return super().handle(entity)


class EpicHandler(BaseHandler):

    def handle(self, entity):
        if "epics".startswith(entity):
            logger.debug("Handling epics")
            return Epic, "epics"
        return super().handle(entity)


class EpicTaskHandler(BaseHandler):

    def handle(self, entity):
        if entity == "epic_tasks":
            logger.debug("Handling epic task")
            return EpicTask, "epic_tasks"
        return super().handle(entity)


class StoryHandler(BaseHandler):

    def handle(self, entity):
        if entity in ("story", "stories"):
            logger.debug("Handling stories")
            return Story, "stories"
        return super().handle(entity)


class StoryTaskHandler(BaseHandler):

    def handle(self, entity):
        if entity == "story_tasks":
            logger.debug("Handling epic task")
            return StoryTask, "story_tasks"
        return super().handle(entity)


class SprintNoteHandler(BaseHandler):

    def handle(self, entity):
        if entity == "notes":
            logger.debug("Handling sprint note")
            return SprintNote, "notes"
        return super().handle(entity)


class NoteHandler(BaseHandler):

    def handle(self, entity):
        if "note".startswith(entity):
            logger.debug("Handling note")
            return BaseNote, "note"
        return super().handle(entity)


class WorkspaceHandler(BaseHandler):

    def handle(self, entity):
        if entity in "workspaces":
            logger.debug("Handling workspace")
            return Workspace, "workspace"
        return super().handle(entity)


class CommandHandler:

    def __init__(self, repo: AbsRepository):
        self.repo = repo
        self.handler = self._init_handlers()
        self.home_dir = os.path.expanduser('~')
        self.console = Console()
        self.config = self._read_config()
        self.printer = printer.Printer(repo=repo, config=self.config)

    def _init_handlers(self):
        handler_chain = BaseHandler(None)
        for handler in [
                TaskHandler, ProjectHandler, UserHandler, TagHandler,
                SprintHandler, SprintTaskHandler, TimeTrackerHandler,
                EpicHandler, StoryHandler, SprintNoteHandler, NoteHandler,
                WorkspaceHandler
        ]:
            new_handler = handler(handler_chain)
            handler_chain = new_handler
        return handler_chain

    def handle(self, entity):
        return self.handler.handle(entity)

    def execute(self,
                command,
                entity_type=None,
                kwargs: Dict[str, Any] = None):
        session = self.repo.session
        if entity_type:
            entity, entity_type = self.handle(entity_type)
        else:
            entity = None
        if not entity and command not in ("calendar", "help"):
            raise ValueError(f"Entity *{entity_type}* is not a valid entity")
        command = format_command(command)
        if command == "list":
            if kwargs.pop("no-expand", None):
                expand_table = False
            else:
                _ = kwargs.pop("expand", None)
                expand_table = True
            print_options = printer.PrintOptions(
                columns=kwargs.pop("columns", ""),
                expand_table=expand_table,
                show_tasks=False,
                show_history=bool(kwargs.pop("show_history", False)),
                show_commentaries=bool(kwargs.pop("show_commentaries", False)),
                show_completed=bool(kwargs.pop("all", False)))
            if entity_type == "notes":
                note_type = get_note_type(kwargs)
                if note_type:
                    notes = self.repo.list(note_type, kwargs)
                    self.printer.print_entities(notes,
                                                "notes",
                                                self.repo,
                                                custom_sort=None,
                                                print_options=print_options)
                else:
                    for note_type in (ProjectNote, TaskNote, SprintNote,
                                      EpicNote, StoryNote):
                        if notes := self.repo.list(note_type, {}):
                            self.console.print(note_type.__name__)
                            self.printer.print_entities(
                                notes,
                                "notes",
                                self.repo,
                                custom_sort=None,
                                print_options=print_options)

                exit()
            if entity_type == "tasks":
                if print_options.show_completed:
                    kwargs[
                        "status"] = "BACKLOG,TODO,IN_PROGRESS,REVIEW,DONE,DELETED"
                elif not kwargs.get("status"):
                    kwargs["status"] = "BACKLOG,TODO,IN_PROGRESS,REVIEW"
            if entity_type in ("stories", "epics"):
                if print_options.show_completed:
                    kwargs["status"] = "ACTIVE,COMPLETED"
                else:
                    kwargs["status"] = "ACTIVE"
            if entity_type == "sprints":
                if print_options.show_completed:
                    kwargs["status"] = "PLANNED,ACTIVE,COMPLETED"
                else:
                    kwargs["status"] = "PLANNED,ACTIVE"
            if entity_type == "projects":
                if print_options.show_completed:
                    kwargs["status"] = "DELETED,ACTIVE,ON_HOLD,COMPLETED"
                    show_completed = True
                else:
                    kwargs["status"] = "ACTIVE"

            custom_sort = kwargs.pop("sort", None)
            if entity_type == "projects" and "project_id" in kwargs:
                kwargs["id"] = kwargs.pop("project_id")
            if has_collaborators := "collaborators" in kwargs:
                collaborator_name = kwargs["collaborators"]
                collaborator = self.repo.list(User,
                                              {"name": collaborator_name})
                if collaborator:
                    task_collaborators = self.repo.list(
                        TaskCollaborator, {"collaborator": collaborator[0].id})
                    if task_collaborators:
                        tasks_with_collaborators = set([
                            task_collaborator.task
                            for task_collaborator in task_collaborators
                        ])
                        del kwargs["collaborators"]
                    else:
                        self.console.print(
                            f"[red]No task with collaborator '{colaborator_name}' found![/red]"
                        )
                        exit()
                else:
                    self.console.print(
                        f"[red]No user '{collaborator_name}' found![/red]")
                    exit()
            else:
                tasks_with_collaborators = None
            if has_tags := "tags" in kwargs:
                tag_text = kwargs["tags"]
                tag = self.repo.list(BaseTag, {"text": tag_text})
                if tag:
                    if entity_type == "tasks":
                        entity_tags = self.repo.list(TaskTag,
                                                     {"tag": tag[0].id})
                    if entity_type == "projects":
                        entity_tags = self.repo.list(ProjectTag,
                                                     {"tag": tag[0].id})
                    if entity_tags:
                        if entity_type == "tasks":
                            entities_with_tag = set(
                                [tag.task for tag in entity_tags])
                            del kwargs["tags"]
                        if entity_type == "projects":
                            entities_with_tag = set(
                                [tag.project for tag in entity_tags])
                            del kwargs["tags"]
                    else:
                        self.console.print(
                            f"[red]No tasks with tag '{tag_text}' found![/red]"
                        )
                        exit()
                else:
                    self.console.print(
                        f"[red]No tag '{tag_text}' found![/red]")
                    exit()
            else:
                entities_with_tag = None
            if entities_with_tag and tasks_with_collaborators:
                filtered_tasks = entities_with_tag.intersection(
                    tasks_with_collaborators)
            elif entities_with_tag:
                filtered_tasks = entities_with_tag
            elif tasks_with_collaborators:
                filtered_tasks = tasks_with_collaborators
            else:
                filtered_tasks = None
            if filtered_tasks:
                kwargs["id"] = ",".join(
                    [str(task_id) for task_id in list(filtered_tasks)])
            elif has_tags or has_collaborators:
                exit("No tasks found")
            entities = self.repo.list(entity, kwargs)
            if custom_sort == "due_date":
                entities_with_due_date = []
                entities_without_due_date = []
                for entity in entities:
                    if entity.due_date:
                        entities_with_due_date.append(entity)
                entities_with_due_date.sort(key=lambda c: c.due_date,
                                            reverse=False)
                entities = entities_with_due_date
            session.commit()
            self.printer.print_entities(entities,
                                        entity_type,
                                        self.repo,
                                        custom_sort,
                                        print_options=print_options)
            logger.info("<list> %s", entity_type)
            return entities, None, None
        elif command == "help":
            print("""
            available commands: 'list', 'show', 'create', 'update', 'done', 'calendar', 'log', 'edit'
            available entities: 'tasks', 'projects', 'commentaries'
            """)
        elif command == "get":
            if entity_type == "tasks" and "status" not in kwargs:
                kwargs["status"] = "BACKLOG,TODO,IN_PROGRESS,REVIEW,DONE"
            if (custom_sort := kwargs.get("sort")):
                del kwargs["sort"]
            else:
                custom_sort = None
            if entity_type == "projects" and "project_id" in kwargs:
                kwargs["id"] = kwargs.pop("project_id")
            entities = self.repo.list(entity, kwargs)
            if custom_sort == "due_date":
                entities_with_due_date = []
                entities_without_due_date = []
                for entity in entities:
                    if entity.due_date:
                        entities_with_due_date.append(entity)
                entities_with_due_date.sort(key=lambda c: c.due_date,
                                            reverse=False)
                entities = entities_with_due_date
            session.commit()
            return entities
        elif command == "calendar":
            kwargs["sort"] = "due_date"
            self.execute("list", "tasks", kwargs)
        elif command == "count":
            entities = self.repo.list(entity, kwargs)
            print(len(entities))
            logger.info("<count> tasks")
        elif command == "collaborate":
            if entity_type not in ("tasks", "projects"):
                raise ValueError(
                    "You can collaborate only on tasks and projects")
            if entity_type == "tasks":
                task_ids = get_ids(kwargs["id"])
                for task_id in task_ids:
                    existing_task = self.repo.list(Task, {"id": task_id})
                    if not existing_task:
                        raise ValueError(
                            f"Task with id {task_id} is not found!")
                    user = self.repo.list(User, {"name": kwargs["name"]})
                    if not user:
                        user = User(**kwargs)
                        self.repo.add(user)
                        session.commit()
                        user_id = user.id
                    else:
                        user_id = user[0].id
                    existing_task_collaborator = self.repo.list(
                        TaskCollaborator, {
                            "task": task_id,
                            "collaborator": user_id
                        })
                    if not existing_task_collaborator:
                        task_ids = get_ids(task_id)
                        for task_id in task_ids:
                            obj = TaskCollaborator(id=task_id,
                                                   collaborator_id=user_id)
                            self.repo.add(obj)
            elif entity_type == "projects":
                project_ids = get_ids(kwargs["id"])
                for project_id in project_ids:
                    existing_project = self.repo.list(Project,
                                                      {"id": project_id})
                    if not existing_project:
                        raise ValueError(
                            f"Project with id {project_id} is not found!")
                    user = self.repo.list(User, {"name": kwargs["name"]})
                    if not user:
                        user = User(**kwargs)
                        self.repo.add(user)
                        session.commit()
                        user_id = user.id
                    else:
                        user_id = user[0].id
                    existing_project_collaborator = self.repo.list(
                        ProjectCollaborator, {
                            "project": project_id,
                            "collaborator": user_id
                        })
                    if not existing_project_collaborator:
                        project_ids = get_ids(project_id)
                        for project_id in project_ids:
                            obj = ProjectCollaborator(id=project_id,
                                                      collaborator_id=user_id)
                            self.repo.add(obj)
            session.commit()
        elif command == "tag":
            if entity_type == "tasks":
                task_id = kwargs["id"]
                existing_task = self.repo.list(Task, {"id": task_id})
                if not existing_task:
                    raise ValueError(f"Task with id {task_id} is not found!")
                tag_text = kwargs.get("text") or kwargs.get("tags")
                tag_info = {"text": tag_text}
                tag = self.repo.list(BaseTag, tag_info)
                if not tag:
                    tag = BaseTag(**tag_info)
                    self.repo.add(tag)
                    session.commit()
                    tag_id = tag.id
                else:
                    tag_id = tag[0].id
                existing_task_tag = self.repo.list(TaskTag, {
                    "task": task_id,
                    "tag": tag_id
                })
                if not existing_task_tag:
                    task_ids = get_ids(task_id)
                    for task_id in task_ids:
                        obj = TaskTag(id=task_id, tag_id=tag_id)
                        self.repo.add(obj)
                    session.commit()
            elif entity_type == "projects":
                project_id = kwargs["id"]
                existing_project = self.repo.list(Project, {"id": project_id})
                if not existing_project:
                    raise ValueError(
                        f"Project with id {project_id} is not found!")
                tag = self.repo.list(BaseTag, {"text": kwargs["text"]})
                if not tag:
                    tag = BaseTag(**kwargs)
                    self.repo.add(tag)
                    session.commit()
                    tag_id = tag.id
                else:
                    tag_id = tag[0].id

                existing_project_tag = self.repo.list(ProjectTag, {
                    "project": project_id,
                    "tag": tag_id
                })
                if not existing_project_tag:
                    project_ids = get_ids(project_id)
                    for project_id in project_ids:
                        obj = ProjectTag(id=project_id, tag_id=tag_id)
                        self.repo.add(obj)
                    session.commit()
        elif command == "note":
            if entity_type == "tasks":
                existing_task = self.repo.list(Task, {"id": kwargs["id"]})
                if not existing_task:
                    raise ValueError(
                        f"Task with id {kwargs['id']} is not found!")
                existing_note = self.repo.list(TaskNote,
                                               {"task": kwargs["id"]})
                if not existing_note:
                    obj = TaskNote(**kwargs)
                    self.repo.add(obj)
                    self.repo.update(Task, kwargs["id"],
                                     {"modification_date": datetime.now()})
                    session.commit()
                else:
                    from rich.markdown import Markdown
                    self.console.print(Markdown(existing_note[0].text))
                    exit()
            elif entity_type == "projects":
                existing_project = self.repo.list(Project,
                                                  {"id": kwargs["id"]})
                if not existing_project:
                    raise ValueError(
                        f"Project with id {kwargs['id']} is not found!")
                obj = ProjectNote(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "epics":
                epic = self.repo.list(Epic, {"id": kwargs["id"]})
                if not epic:
                    raise ValueError(
                        f"Epic with id {kwargs['id']} is not found!")
                obj = EpicNote(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "stories":
                epic = self.repo.list(Story, {"id": kwargs["id"]})
                if not epic:
                    raise ValueError(
                        f"Story with id {kwargs['id']} is not found!")
                obj = StoryNote(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "sprints":
                sprint = self.repo.list(Sprint, {"id": kwargs["id"]})
                if not sprint:
                    raise ValueError(
                        f"Sprint with id {kwargs['id']} is not found!")
                obj = SprintNote(**kwargs)
                self.repo.add(obj)
                session.commit()
            self.printer.print_new_object(obj)
        elif command == "comment":
            if entity_type == "tasks":
                existing_task = self.repo.list(Task, {"id": kwargs["id"]})
                if not existing_task:
                    raise ValueError(
                        f"Task with id {kwargs['id']} is not found!")
                obj = TaskCommentary(**kwargs)
                self.repo.add(obj)
                self.repo.update(Task, kwargs["id"],
                                 {"modification_date": datetime.now()})
                session.commit()
            elif entity_type == "projects":
                existing_project = self.repo.list(Project,
                                                  {"id": kwargs["id"]})
                if not existing_project:
                    raise ValueError(
                        f"Project with id {kwargs['id']} is not found!")
                obj = ProjectCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "epics":
                epic = self.repo.list(Epic, {"id": kwargs["id"]})
                if not epic:
                    raise ValueError(
                        f"Epic with id {kwargs['id']} is not found!")
                obj = EpicCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "stories":
                epic = self.repo.list(Story, {"id": kwargs["id"]})
                if not epic:
                    raise ValueError(
                        f"Story with id {kwargs['id']} is not found!")
                obj = StoryCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "sprints":
                sprint = self.repo.list(Sprint, {"id": kwargs["id"]})
                if not sprint:
                    raise ValueError(
                        f"Sprint with id {kwargs['id']} is not found!")
                obj = SprintCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
            self.printer.print_new_object(obj)
        elif command == "delete":
            if entity not in (Task, ):
                raise ValueError("'delete' operation only allowed for tasks!")
            task_ids = get_ids(kwargs.get("id"))
            for task_id in task_ids:
                if not (task := self.repo.get_by_id(Task, task_id)):
                    exit(f"Task id {task_id} is not found")
                else:
                    if all(key == "id" for key in kwargs.keys()):
                        deletion_reason = input(
                            f"Provide brief explanation why you're deleting task <{task_id}>: "
                        )
                        if deletion_reason:
                            self.execute("comment", "tasks", {
                                "id": task_id,
                                "text": deletion_reason
                            })
                        kwargs.update({"status": "DELETED"})
                        self.execute("update", entity_type, kwargs)
                if epic_id := kwargs.get("epic_id"):
                    epic = self.repo.list(Epic, {"id": epic_id})
                    if not epic:
                        exit(f"Epic id {epic_id} is not found")
                    if not (epic_task := self.repo.list(
                            EpicTask, {
                                "task": task_id,
                                "epic": epic_id
                            })):
                        exit("task is not in epic")
                    self.repo.delete(EpicTask, epic_task[0].id)
                if story_id := kwargs.get("story_id"):
                    story = self.repo.list(Story, {"id": story_id})
                    if not story:
                        exit(f"Story id {story_id} is not found")
                    if not (story_task := self.repo.list(
                            StoryTask, {
                                "task": task_id,
                                "story": story_id
                            })):
                        exit("task is not in story")
                    self.repo.delete(StoryTask, story_task[0].id)
                if sprint_id := kwargs.get("sprint_id"):
                    if isinstance(sprint_id, bool):
                        sprint = get_active_sprint(self.repo)
                    else:
                        sprint = self.repo.get_by_id(Sprint, sprint_id)
                    if not sprint:
                        exit(f"Sprint id {sprint_id} is not found")
                    if sprint.status.name == "COMPLETED":
                        exit("Cannot add task to a finished sprint")
                    if not (sprint_task := self.repo.list(
                            SprintTask, {
                                "task": task_id,
                                "sprint": sprint.id
                            })):
                        exit("task is not in sprint")
                    self.repo.delete(SprintTask, sprint_task[0].id)
            session.commit()
        elif command == "connect":
            if entity not in (Task, Project):
                raise ValueError(
                    "'connect' operation only allowed for tasks and project!")
            task_ids = get_ids(kwargs.get("id"))
            for task_id in task_ids:
                if entity_type == "projects":
                    asana_project_id = kwargs.get("external_project")
                    asana_project = self.repo.get_by_id(AsanaProject, task_id)
                    if not asana_project:
                        obj = AsanaProject(id=task_id,
                                           asana_project_id=asana_project_id)
                        self.repo.add(obj)
                    else:
                        self.repo.update(
                            AsanaProject, task_id,
                            {"asana_project_id": asana_project_id})
                elif entity_type == "tasks":
                    asana_task_id = kwargs.get("external_task")
                    asana_task = self.repo.get_by_id(AsanaProject, task_id)
                    if not asana_task:
                        obj = AsanaTask(
                            project=task.project,
                            id=task_id,
                            asana_task_id=kwargs.get("external_task"))
                        self.repo.add(obj)
                    else:
                        self.repo.update(AsanaTask, task_id,
                                         {"asana_task_id": asana_task_id})
            self.repo.session.commit()
        elif command == "add":
            if entity not in (Task, Story, Epic):
                raise ValueError(
                    "'add' operation only allowed for tasks, epics, and stories!"
                )
            task_ids = get_ids(kwargs.get("id"))
            for task_id in task_ids:
                if not (added_task := self.repo.get_by_id(entity, task_id)):
                    exit(
                        f"{entity.__class__.__name__} id {task_id} is not found"
                    )

                if epic_id := kwargs.get("epic_id"):
                    epic = self.repo.list(Epic, {"id": epic_id})
                    if not epic:
                        raise EntityNotFound(f"Epic id {epic_id} is not found")
                    obj = EpicTask(task=task_id, epic=epic_id)
                    if self.repo.list(EpicTask, {
                            "task": obj.task,
                            "epic": obj.epic
                    }):
                        raise TaskAddedToEntity("task already added to epic")
                    self.repo.add(obj)
                    self.execute("tag", "tasks", {
                        "id": obj.task,
                        "tags": f"epic:{obj.epic}"
                    })
                if story_id := kwargs.get("story_id"):
                    story = self.repo.list(Story, {"id": story_id})
                    if not story:
                        raise EntityNotFound(
                            f"Story id {story_id} is not found")
                    obj = StoryTask(task=task_id, story=story_id)
                    if self.repo.list(StoryTask, {
                            "task": obj.task,
                            "story": obj.story
                    }):
                        raise TaskAddedToEntity("task already added to story")
                    self.repo.add(obj)
                    self.execute("tag", "tasks", {
                        "id": obj.task,
                        "tags": f"story:{obj.story}"
                    })
                if sprint_id := kwargs.get("sprint_id"):
                    if isinstance(sprint_id, bool):
                        sprint = get_active_sprint(self.repo)
                    else:
                        sprint = self.repo.get_by_id(Sprint, sprint_id)
                    if not sprint:
                        raise EntityNotFound(
                            f"Sprint id {sprint.id} is not found")
                    if sprint.status.name == "COMPLETED":
                        raise TaskAddedToCompletedEntity(
                            "Cannot add task to a finished sprint")
                    if entity in (Task, ):
                        self._add_task_to_sprint(added_task, sprint)
                    else:
                        for entity_task in added_task.tasks:
                            if entity_task.tasks.status.name not in (
                                    "DONE", "DELETED"):
                                self._add_task_to_sprint(
                                    entity_task.tasks, sprint)
                if story_points := kwargs.get("story_points"):
                    sprint_task = self.execute("get", "sprint_tasks",
                                               {"task": task_id})
                    if not sprint_task:
                        exit(f"Task id {task_id} is not part of any sprint")
                    elif len(sprint_task) > 1:
                        sprint_ids = [str(task.sprint) for task in sprint_task]
                        sprint_id = input(
                            f"Task task_id is part of several sprints, "
                            "where do you want to assign story points? "
                            f"(sprints: {', '.join(sprint_ids)}): ")
                        if sprint_id not in sprint_ids:
                            exit("Incorrect sprint id")
                        sprint_task_id = [
                            task.id for task in sprint_task
                            if task.sprint == int(sprint_id)
                        ][0]

                    else:
                        sprint_task_id = sprint_task[0].id
                    self.repo.update(SprintTask, sprint_task_id,
                                     {"story_points": story_points})
            session.commit()
        elif command == "create":
            kwargs["created_by"] = services.lookup_user_id(
                self.config.get("user"), self.repo)
            if entity == BaseNote:
                if "task_id" in kwargs:
                    create_dict = note_edit_vim(kwargs.pop("task_id"), "task",
                                                kwargs)
                    create_dict.update(kwargs)
                    obj = TaskNote(**create_dict)
                elif "project" in kwargs:
                    create_dict = note_edit_vim(kwargs.pop("project"),
                                                "project", kwargs)
                    create_dict.update(kwargs)
                    obj = ProjectNote(**create_dict)
                elif "sprint_id" in kwargs:
                    create_dict = note_edit_vim(kwargs.pop("sprint_id"),
                                                "sprint", kwargs)
                    create_dict.update(kwargs)
                    obj = SprintNote(**create_dict)
                elif "epic_id" in kwargs:
                    create_dict = note_edit_vim(kwargs.pop("epic_id"), "epic",
                                                kwargs)
                    create_dict.update(kwargs)
                    obj = EpicNote(**create_dict)
                elif "story_id" in kwargs:
                    create_dict = note_edit_vim(kwargs.pop("story_id"),
                                                "story", kwargs)
                    create_dict.update(kwargs)
                    obj = StoryNote(**create_dict)
                self.repo.add(obj)
                session.commit()
                exit()
            else:
                if len(kwargs) == 1:
                    message_template = generate_message_template(entity=entity)
                    with tempfile.NamedTemporaryFile(suffix=".tmp") as tf:
                        tf.write(message_template.encode("utf-8"))
                        tf.flush()
                        run(["vim", "+2", tf.name])
                        tf.seek(0)
                        new_entry = tf.read()

                    updated_entry = {}
                    commentary = {}
                    collaborators = {}
                    tags = {}
                    new_entry = new_entry.decode("utf-8").rstrip()
                    for i, row in enumerate(new_entry.split("\n")):
                        if not row.startswith("#") and row:
                            if row.startswith("--"):
                                continue
                            entry_type, *entry_value = row.split(": ",
                                                                 maxsplit=2)
                            entry_value = ": ".join(entry_value)
                            entry_value = entry_value.strip()
                            try:
                                if entry_value:
                                    if entry_type == "status":
                                        entry_value = convert_status(
                                            entry_value)
                                    if entry_type.endswith("date"):
                                        entry_value = datetime.strptime(
                                            entry_value, "%Y-%m-%d")
                                    updated_entry[entry_type] = entry_value
                            except TerkaException as e:
                                pass

                else:
                    updated_entry = {}
                if updated_entry:
                    kwargs.update(updated_entry)
                    if "due_date" in kwargs:
                        kwargs["due_date"] = convert_date(kwargs["due_date"])
                    kwargs = update_task_dict(kwargs, self.repo)
                obj = entity(**kwargs)
            if entity in (Task, Project, BaseNote):
                if (existing_obj := self.repo.list(entity,
                                                   {"name": obj.name})):
                    print("Found existing entity\n")
                    existing_id = str(existing_obj[0].id)
                    self.execute("show", entity_type, {"id": existing_id})
                    answer = input(
                        "Do you want to create entity anyway? [Y/n] ")
                    if answer.lower() != "y":
                        exit()
            self.repo.add(obj)
            session.commit()
            if tag_text := kwargs.get("tags"):
                self.execute("tag", entity_type, {
                    "id": obj.id,
                    "text": tag_text
                })
            if collaborator_name := kwargs.get("collaborators"):
                self.execute("collaborate", entity_type, {
                    "id": obj.id,
                    "name": collaborator_name
                })
            if hasattr(obj, "project") and obj.project:
                project = services.lookup_project_name(obj.project, self.repo)
            else:
                project = None
            if epic_id := kwargs.get("epics"):
                self.execute("add", entity_type, {
                    "id": obj.id,
                    "epic_id": epic_id
                })
            if sprint_id := kwargs.get("sprints"):
                self.execute("add", entity_type, {
                    "id": obj.id,
                    "sprint_id": sprint_id
                })
            if story_id := kwargs.get("stories"):
                self.execute("add", entity_type, {
                    "id": obj.id,
                    "story_id": story_id
                })
            self.printer.print_new_object(obj, project)
            logger.info("<create> %s: %s", entity_type, obj.id)
            return entity, obj
        elif command == "update":
            if (task_id := kwargs.get("id")):
                del kwargs["id"]
                tasks = get_ids(task_id)
                for task in tasks:
                    if (old_values := self.repo.list(entity, {"id": task})):
                        old_values = old_values[0]
                        old_settings = {
                            key: getattr(old_values, key)
                            for key in kwargs.keys()
                        }
                        if old_settings != kwargs:
                            new_kwargs = {}
                            for k, v in kwargs.items():
                                if old_settings[k] != v:
                                    new_kwargs[k] = v
                            self.repo.update(entity, task, new_kwargs)
                            logger.info("<update> %s: %s", entity_type, task)
                            now = datetime.now()
                            if entity_type in ("tasks", "projects"):
                                for key, value in new_kwargs.items():
                                    try:
                                        old_value = old_settings[key].name
                                    except:
                                        old_value = old_settings[key]
                                    if entity_type == "projects":
                                        event = ProjectEvent
                                    elif entity_type == "tasks":
                                        event = TaskEvent
                                    if old_value != value:
                                        update_event = event(
                                            task, key, old_value, value, now)
                                        self.repo.add(update_event)
                                        self.printer.print_new_object(
                                            update_event)
                            if hasattr(entity, "modification_date"):
                                self.repo.update(entity, task,
                                                 {"modification_date": now})
                        else:
                            print(
                                "No changes were proposed to the existing entity"
                            )
                    session.commit()
                return entity, None
            else:
                raise ValueError("No task_id is provided")
        elif command == "report":
            start_date = kwargs.pop("start_date", get_days_ago(7))
            end_date = kwargs.pop("end_date", get_days_ago(1))
            show_tasks = kwargs.pop("tasks", False)

            created_tasks = self.repo.session.query(Task).filter(
                Task.creation_date >= start_date,
                Task.creation_date <= end_date).all()
            completed_task_events = self.repo.session.query(TaskEvent).filter(
                TaskEvent.date >= start_date, TaskEvent.date <= end_date,
                TaskEvent.new_value == "DONE").all()
            completed_task_ids = ",".join(
                list(set([str(task.task) for task in completed_task_events])))
            completed_tasks = self.execute("get", "tasks",
                                           {"id": completed_task_ids})
            if show_tasks:
                self.printer.print_entities(created_tasks,
                                            "tasks",
                                            self.repo,
                                            custom_sort="id")
                self.printer.print_entities(
                    completed_tasks,
                    "tasks",
                    self.repo,
                    custom_sort="id",
                    print_options=printer.PrintOptions(show_completed=True))
            self.console.print(
                f"For the period: {start_date.strftime('%Y-%m-%d')} - {end_date.strftime('%Y-%m-%d')}"
            )
            self.console.print(f"{len(created_tasks)} tasks were created")
            self.console.print(f"{len(completed_tasks)} tasks were completed")
        elif command == "edit":
            if (task_id := kwargs.get("id")):
                tasks = get_ids(task_id)
                for task in tasks:
                    if entity_type == "note":
                        note_type = get_note_type(kwargs)
                        note = self.repo.get_by_id(note_type, task)
                        if note:
                            with tempfile.NamedTemporaryFile(
                                    suffix=".md") as tf:
                                if isinstance(note.text, (bytes, bytearray)):
                                    text = note.text
                                else:
                                    text = note.text.encode("utf-8")
                                tf.write(text)
                                tf.flush()
                                run(["vim", "+2", tf.name])
                                tf.seek(0)
                                new_entry = tf.read()
                            self.repo.update(note_type, task,
                                             {"text": new_entry})
                            session.commit()
                            exit()
                        else:
                            print(f"There's no note with id {task}")

                    if task.isdigit():
                        entities = self.repo.get_by_id(entity, task)
                        task_id = task
                    else:
                        entities = self.repo.get(entity, task)
                        task_id = entities.id
                    if entity_type in ("tasks", "projects", "sprints", "epics",
                                       "stories"):
                        task = entities
                        # if kwargs.get("description"):
                        #     current_entry = task.description
                        #     current_type = "description"
                        # elif kwargs.get("name"):
                        #     current_entry = task.name
                        #     current_type = "name"
                        # elif kwargs.get("goal"):
                        #     current_entry = task.goal
                        #     current_type = "goal"
                        # else:
                        #     raise ValueError(
                        #         "Either name or description should be specified!"
                        #     )

                    message_template = generate_message_template(
                        task, self.repo, kwargs)
                    with tempfile.NamedTemporaryFile(suffix=".tmp") as tf:
                        tf.write(message_template.encode("utf-8"))
                        tf.flush()
                        run(["vim", "+2", tf.name])
                        tf.seek(0)
                        new_entry = tf.read()

                    updated_entry = {}
                    time_entry = {}
                    commentary = {}
                    collaborators = {}
                    tags = {}
                    new_entry = new_entry.decode("utf-8").rstrip()
                    for i, row in enumerate(new_entry.split("\n")):
                        if not row.startswith("#") and row:
                            if row.startswith("--"):
                                continue
                            entry_type, *entry_value = row.split(": ",
                                                                 maxsplit=2)
                            entry_value = ": ".join(entry_value)
                            entry_value = entry_value.strip()
                            try:
                                if entry_type == "tags":
                                    self.execute("tag", entity_type, {
                                        "id": task.id,
                                        "tags": entry_value
                                    })
                                elif entry_type == "comment":
                                    self.execute("comment", entity_type, {
                                        "id": task.id,
                                        "text": entry_value
                                    })
                                elif entry_type == "collaborators":
                                    self.execute("collaborate", entity_type, {
                                        "id": task.id,
                                        "name": entry_value
                                    })
                                elif entry_type == "epics":
                                    self.execute("add", entity_type, {
                                        "id": task.id,
                                        "epic_id": entry_value
                                    })
                                elif entry_type == "sprints":
                                    self.execute("add", entity_type, {
                                        "id": task.id,
                                        "sprint_id": entry_value
                                    })
                                elif entry_type == "stories":
                                    self.execute("add", entity_type, {
                                        "id": task.id,
                                        "story_id": entry_value
                                    })
                                elif entry_type == "time_spent":
                                    time_spent = entry_value.split(" (")
                                    try:
                                        time_spent = time_spent[0].strip()
                                        time_spent_minutes = int(time_spent)
                                        if time_spent_minutes:
                                            self.execute(
                                                "track", entity_type, {
                                                    "id": task.id,
                                                    "minutes":
                                                    time_spent_minutes
                                                })
                                    except Exception:
                                        pass
                                else:
                                    if entry_value:
                                        if entry_type == "status":
                                            entry_value = convert_status(
                                                entry_value)
                                        if entry_type.endswith("date"):
                                            entry_value = datetime.strptime(
                                                entry_value, "%Y-%m-%d")
                                        updated_entry[entry_type] = entry_value
                            except TerkaException as e:
                                pass
                if updated_entry:
                    updated_entry.update({"id": task.id})
                    # new_kwargs = {
                    #     "id": task_id,
                    #     current_type: " ".join(updated_entry)
                    # }
                    self.execute("update", entity_type, updated_entry)
                return entities, None, None
        elif command == "start":
            sprint_id = kwargs.get("id")
            kwargs.update({"status": "ACTIVE"})
            [sprint] = self.execute("get", "sprints", {"id": sprint_id})
            if sprint.end_date < datetime.today().date():
                self.console.print(
                    "[red]Cannot start the sprint, end date in the past[/red]")
                exit()
            self.execute("update", "sprints", kwargs)
            for sprint_task in sprint.tasks:
                task = sprint_task.tasks
                task_params = {"id": task.id}
                if task.status.name == "BACKLOG":
                    task_params.update({"status": "TODO"})
                if not task.due_date or task.due_date > sprint.end_date:
                    task_params.update({"due_date": sprint.end_date})
                self.execute("update", "tasks", task_params)
                if sprint_task.story_points == 0:
                    story_points = input(
                        f"Please enter story points estimation for task <{task.id}>: {task.name}: "
                    )
                    try:
                        float(story_points)
                    except ValueError:
                        self.console.print(
                            "[red]Provide number when specifying story points[/red]"
                        )
                        story_points = 0
                    self.execute("update", "sprint_tasks", {
                        "id": sprint_task.id,
                        "story_points": story_points
                    })

        elif command == "show":
            if len(kwargs) == 1 and "project" in kwargs:
                kwargs["id"] = kwargs["project"]
                del kwargs["project"]
            print_options = printer.PrintOptions(
                show_history=bool(kwargs.pop("show_history", False)),
                show_commentaries=bool(kwargs.pop("show_commentaries", False)),
                show_completed=bool(kwargs.pop("all", False)),
                show_notes=bool(kwargs.pop("show_notes", False)),
                show_viz=kwargs.pop("show_viz", False))
            if kwargs.pop("partial_project_view", False):
                print_options.show_epics = bool(kwargs.pop("epics", False))
                print_options.show_tasks = bool(kwargs.pop("tasks", False))
                print_options.show_stories = bool(kwargs.pop("stories", False))
            if entity_type == "note":
                note_type = get_note_type(kwargs)
                note = self.repo.get_by_id(note_type, kwargs.get("id"))
                self.printer.print_entity(kwargs.get("id"), "notes", [note],
                                          self.repo, None, print_options)
                exit()
            if entity_type == "tasks":
                print_options.show_completed = True
            if not (task_id := kwargs.get("id")):
                if entity_type == "sprints":
                    active_sprint = get_active_sprint(self.repo)
                    task_id = active_sprint.id
            tasks = get_ids(task_id)
            if "id" in kwargs:
                kwargs.pop("id")
            for task in tasks:
                if task.isdigit():
                    entities = self.repo.list(entity, {"id": task})
                else:
                    entities = self.repo.list(entity, {"name": task})
                self.printer.print_entity(task, entity_type, entities,
                                          self.repo, print_options, kwargs)
                logger.info("<show> %s: %s", entity_type, task_id)
            return entities, None, None
        elif command == "done":
            if entity_type not in ("tasks", "sprints", "stories", "epics"):
                raise ValueError("can complete only tasks and sprints")
            elif entity_type == "tasks":
                kwargs.update({"status": "DONE"})
                self.execute("edit", entity_type, kwargs)
                self.console.print(
                    "[green]Yay! You've just completed a task![/green]")
            elif entity_type in ("sprints", "epics", "stories"):
                kwargs.update({"status": "COMPLETED"})
                [entity] = self.execute("get", entity_type,
                                        {"id": kwargs.get("id")})

                if entity_type == "sprints":
                    entity.complete(entity.tasks)
                    for entity_task in entity.tasks:
                        task_status = entity_task.tasks.status.name
                        if task_status not in ("DONE", "DELETED"):
                            update_dict = {
                                "id": entity_task.task,
                                "status": "BACKLOG"
                                if task_status == "TODO" else task_status,
                                "due_date": None
                            }
                            self.execute("update", "tasks", update_dict)
                else:
                    entity.complete(entity.tasks)
                    for entity_task in entity.tasks:
                        self.execute("update", "tasks", {
                            "id": entity_task.task,
                            "status": "DONE"
                        })
                self.execute("update", entity_type, kwargs)
        elif command == "track":
            if entity_type != "tasks":
                raise ValueError("can track only tasks")
            if "hours" in kwargs and "minutes" in kwargs:
                exit("specify only -H (hours) or -M (minutes) value")
            if "hours" in kwargs or "minutes" in kwargs:
                if "hours" in kwargs:
                    kwargs["time_spent_minutes"] = float(kwargs["hours"]) * 60
                    del kwargs["hours"]
                else:
                    kwargs["time_spent_minutes"] = float(kwargs["minutes"])
                    del kwargs["minutes"]
                kwargs["task"] = kwargs["id"]
                del kwargs["id"]
                self.execute("create", "time_entry", kwargs)
            else:
                exit("tracking missing -H (hours) or -M (minutes) value")
        else:
            raise ValueError(f"Uknown command: {command}")

    def _add_task_to_sprint(self, task, sprint):
        if self.repo.list(SprintTask, {"task": task.id, "sprint": sprint.id}):
            raise TaskAddedToEntity("task already added to sprint")
        if sprint.status.name == "ACTIVE":

            story_points = input(
                f"Please enter story points estimation for task <{task.id}>: {task.name}: "
            )
            try:
                float(story_points)
            except ValueError:
                self.console.print(
                    "[red]Provide number when specifying story points[/red]")
                exit()
        else:
            story_points = 0
        obj = SprintTask(task=task.id,
                         story_points=story_points,
                         sprint=sprint.id,
                         is_active_link=True)
        self.repo.add(obj)
        sprint_task_id = obj.task
        if sprint_task_id and sprint.status.name == "ACTIVE":
            task_params = {"id": task.id}
            if task.status.name == "BACKLOG":
                task_params.update({"status": "TODO"})
            if not task.due_date or task.due_date > sprint.end_date:
                task_params.update({"due_date": sprint.end_date})
            self.execute("update", "tasks", task_params)
            self.execute("tag", "tasks", {
                "id": task.id,
                "tags": f"sprint:{sprint.id}"
            })

    def _read_config(self) -> Dict[str, Any]:
        with open(f"{self.home_dir}/.terka/config.yaml", "r",
                  encoding="utf-8") as f:
            config = yaml.safe_load(f)
        return config


def get_ids(ids: Union[str, int]) -> List[Union[int, str]]:
    id_string = str(ids)
    if "," in id_string:
        ids = id_string.split(",")
    elif ".." in id_string:
        task_range = id_string.split("..")
        ids = range(int(task_range[0]), int(task_range[1]) + 1)
    else:
        ids = [id_string]
    return ids


def note_edit_vim(entity_id: str, attached_entity: str,
                  kwargs: Dict[str, Any]) -> Dict[str, Any]:
    message_template = f"---You are adding note for {attached_entity} {entity_id}"
    with tempfile.NamedTemporaryFile(suffix=".md") as tf:
        if not isinstance(message_template, (bytes, bytearray)):
            message_template = message_template.encode("utf-8")
        tf.write(message_template)
        tf.flush()
        run(["vim", "+2", tf.name])
        tf.seek(0)
        new_entry = tf.read()
    text = []
    new_entry = new_entry.decode("utf-8").rstrip()
    for i, row in enumerate(new_entry.split("\n")):
        if row.startswith("---"):
            continue
        if row:
            text.append(row)
    if "name" not in kwargs:
        name = input("Would you like to enter the name for the note? ")
    create_dict = {"id": entity_id, "text": "\n".join(text), "name": name}
    return create_dict


def get_note_type(kwargs: Dict[str, Any]) -> BaseNote:
    if "project" in kwargs:
        return ProjectNote
    if "task_id" in kwargs:
        return TaskNote
    if "sprint_id" in kwargs:
        return SprintNote
    if "story_id" in kwargs:
        return StoryNote
    if "epic_id" in kwargs:
        return EpicNote


def get_active_sprint(repo: AbsRepository) -> int:
    active_sprint = repo.list(Sprint, {"status": "ACTIVE"})
    if len(active_sprint) == 1:
        return active_sprint[0]
    elif len(active_sprint) > 1:
        exit("More than 1 active sprint, please specify the sprint_id")
    else:
        exit("No active sprints")


def get_days_ago(lookback: int = 0) -> datetime.date:
    return datetime.now().date() - timedelta(days=lookback)



import terka
from textual.screen import Screen
from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, Label, Tabs, DataTable, TabbedContent, TabPane, Static

EPICS = [
    ("id", "name", "description"),
    (1, "First", None),
    (2, "Second", None),
    (3, "Third", "Sample descr"),
]

TASKS = [
    ("id", "name", "description"),
    (10, "First", None),
    (20, "Second", None),
    (30, "Third", "Sample descr"),
]

STORIES = [
    ("id", "name", "description"),
    (10, "First", None),
    (20, "Second", None),
    (30, "Third", "Sample descr"),
]


class TerkaProject(Screen):
    """Demonstrates the Tabs widget."""

    CSS = """
    Tabs {
        dock: top;
    }
    Screen {
        align: center top;
    }
    .header {
        margin:1 1;
        width: 100%;
        height: 5%;
        background: $panel;
        border: tall $primary;
        content-align: center middle;
    }
    """

    BINDINGS = [("e", "epics", "Epics"), ("t", "tasks", "Tasks"),
                ("s", "stories", "Stories"), ("n", "notes", "Notes"),
                ("o", "overview", "Overview"), ("q", "quit", "Quit")]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static("Project name", classes="header")
        with TabbedContent(initial="tasks"):
            with TabPane("Tasks", id="tasks"):
                table = DataTable(id="tasks")
                table.add_columns(*TASKS[0])
                table.add_rows(TASKS[1:])
                yield table
            with TabPane("Epics", id="epics"):
                table = DataTable(id="epics")
                table.add_columns(*EPICS[0])
                table.add_rows(EPICS[1:])
                yield table
            with TabPane("Stories", id="stories"):
                table = DataTable(id="stories")
                table.add_columns(*STORIES[0])
                table.add_rows(STORIES[1:])
                yield table
            with TabPane("Notes", id="notes"):
                table = DataTable(id="notes")
                table.add_columns(*STORIES[0])
                table.add_rows(STORIES[1:])
                yield table
            with TabPane("Overview", id="overview"):
                table = DataTable(id="overview")
                table.add_columns(*STORIES[0])
                table.add_rows(STORIES[1:])
                yield table
        yield Footer()

    # def on_mount(self) -> None:
    #     """Focus the tabs when the app starts."""
    #     self.query_one(Tabs).focus()

    def action_epics(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "epics"

    def action_tasks(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "tasks"

    def action_stories(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "stories"

    def action_overview(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "overview"

    def action_notes(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "notes"

    # def action_add(self) -> None:
    #     """Add a new tab."""
    #     tabs = self.query_one(Tabs)
    #     # Cycle the names
    #     NAMES[:] = [*NAMES[1:], NAMES[0]]
    #     tabs.add_tab(NAMES[0])

    # def action_remove(self) -> None:
    #     """Remove active tab."""
    #     tabs = self.query_one(Tabs)
    #     active_tab = tabs.active_tab
    #     if active_tab is not None:
    #         tabs.remove_tab(active_tab.id)

    # def action_clear(self) -> None:
    #     """Clear the tabs."""
    #     self.query_one(Tabs).clear()


class TerkaSprint(Screen):
    """Demonstrates the Tabs widget."""

    CSS = """
    Tabs {
        dock: top;
    }
    Screen {
        align: center top;
    }
    .header {
        margin:1 1;
        width: 100%;
        height: 5%;
        background: $panel;
        border: tall $primary;
        content-align: center middle;
    }
    """

    BINDINGS = [("t", "tasks", "Tasks"), ("n", "notes", "Notes"),
                ("o", "overview", "Overview"), ("q", "quit", "Quit")]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static("Sprint name", classes="header")
        with TabbedContent(initial="tasks"):
            with TabPane("Tasks", id="tasks"):
                table = DataTable(id="tasks")
                table.add_columns(*TASKS[0])
                table.add_rows(TASKS[1:])
                yield table
            with TabPane("Overview", id="overview"):
                table = DataTable(id="overview")
                table.add_columns(*STORIES[0])
                table.add_rows(STORIES[1:])
                yield table
        yield Footer()

    # def on_mount(self) -> None:
    #     """Focus the tabs when the app starts."""
    #     self.query_one(Tabs).focus()

    def action_epics(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "epics"

    def action_tasks(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "tasks"

    def action_stories(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "stories"

    def action_overview(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "overview"

    def action_notes(self) -> None:
        """Add a new tab."""
        self.query_one(TabbedContent).active = "notes"

    # def action_add(self) -> None:
    #     """Add a new tab."""
    #     tabs = self.query_one(Tabs)
    #     # Cycle the names
    #     NAMES[:] = [*NAMES[1:], NAMES[0]]
    #     tabs.add_tab(NAMES[0])

    # def action_remove(self) -> None:
    #     """Remove active tab."""
    #     tabs = self.query_one(Tabs)
    #     active_tab = tabs.active_tab
    #     if active_tab is not None:
    #         tabs.remove_tab(active_tab.id)


    # def action_clear(self) -> None:
    #     """Clear the tabs."""
    #     self.query_one(Tabs).clear()
class Terka(App):
    BINDINGS = [
        ("1", "switch_mode('project')", "Dashboard"),
        ("2", "switch_mode('sprint')", "Settings"),
    ]
    MODES = {"project": TerkaProjects, "sprint": TerkaCurrentSprint}

    def on_mount(self) -> None:
        self.switch_mode("project")


if __name__ == "__main__":

    app = Terka()
    app.run()



from dataclasses import dataclass, asdict
from terka.domain.task import TaskInfo


@dataclass
class Event:
    ...


# Base
@dataclass
class Created(Event):
    id: int


@dataclass
class Completed(Event):
    id: int
    comment: str | None = None


@dataclass
class Deleted(Event):
    id: int
    comment: str | None = None


@dataclass
class Updated(Event):
    id: int


@dataclass
class Commented(Event):
    id: int
    text: str



# TASKS
@dataclass
class TaskEvent(Event):
    ...


@dataclass
class SprintEvent(Event):
    ...


@dataclass
class ProjectEvent(Event):
    ...


@dataclass
class TaskCreated(TaskEvent):
    id: int


@dataclass
class TaskCompleted(Completed):
    hours: int | None = None


@dataclass
class TaskDeleted(Deleted):
    ...


@dataclass
class UpdateMask:
    name: str | None = None
    description: str | None = None
    project: str | None = None
    assignee: str | None = None
    due_date: str | None = None
    status: str = "BACKLOG"
    priority: str = "NORMAL"

    def get_only_set_attributes(self) -> dict:
        return {key: value for key, value in asdict(self).items() if value}


@dataclass
class TaskUpdated(TaskEvent):
    id: int
    update_mask: UpdateMask


@dataclass
class TaskCommentAdded(TaskEvent):
    id: int
    text: str


@dataclass
class TaskHoursSubmitted(TaskEvent):
    id: int
    hours: int


@dataclass
class TaskCollaboratorAdded(TaskEvent):
    id: int
    collaborator: str


@dataclass
class TaskTagAdded(TaskEvent):
    id: int
    tag: str


@dataclass
class TaskAddedToSprint(TaskEvent):
    id: int
    sprint_id: int


@dataclass
class TaskAddedToEpic(TaskEvent):
    id: int
    epic_id: int


@dataclass
class TaskAddedToStory(TaskEvent):
    id: int
    story_id: int


# PROJECTS
@dataclass
class ProjectCreated(Created):
    ...

@dataclass
class ProjectCompleted(Completed):
    ...
    
@dataclass
class ProjectDeleted(Deleted):
    ...

@dataclass
class ProjectCommented(Commented):
    ...


@dataclass
class SprintTaskStoryPointAssigned(SprintEvent):
    id: int
    story_points: float


# EPICS
@dataclass
class EpicCreated(Created):
    ...


@dataclass
class EpicCompleted(Completed):
    ...


@dataclass
class EpicDeleted(Deleted):
    ...


@dataclass
class EpicUpdated(Updated):
    ...


@dataclass
class EpicCommented(Commented):
    ...


# STORIES
@dataclass
class StoryCreated(Created):
    ...


@dataclass
class StoryCompleted(Completed):
    ...


@dataclass
class StoryDeleted(Deleted):
    ...


@dataclass
class StoryUpdated(Updated):
    ...


@dataclass
class StoryCommented(Commented):
    ...

# SPRINTS
@dataclass
class SprintCompleted(Completed):
    ...




from datetime import datetime
from dataclasses import dataclass


@dataclass
class AsanaTask:
    id: int
    project: int
    asana_task_id: str
    sync_date: datetime


@dataclass
class AsanaProject:
    id: int
    asana_project_id: str
    sync_date: datetime



from typing import Any, Dict, List, Optional, Tuple
from dataclasses import dataclass
from operator import attrgetter
import inspect
from copy import deepcopy

from collections import defaultdict
import itertools
from datetime import datetime, date, timedelta
import plotext as plt
import rich
import pandas as pd
import numpy as np
import matplotlib.pyplot
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table
from statistics import mean, median

from terka.service_layer import services, views, formatter
from terka.service_layer.ui import TerkaTask, TerkaProject, TerkaSprint


@dataclass
class PrintOptions:
    show_tasks: bool = True
    show_history: bool = False
    show_commentaries: bool = False
    show_completed: bool = False
    show_epics: bool = True
    show_stories: bool = True
    show_notes: bool = True
    show_viz: bool = False
    columns: str = ""
    expand_table: bool = True


class Printer:

    def __init__(self, repo, box=rich.box.SIMPLE, config = None) -> None:
        self.console = Console()
        self.box = box
        self.repo = repo
        self.config = config

    def print_new_object(self, obj, project=None):
        table = Table(box=self.box)
        attributes = self._get_attributes(obj)
        for column, value in attributes:
            if value:
                table.add_column(column)
        table.add_row(*list(zip(*attributes))[1])
        if table.row_count:
            self.console.print(f"Added new {obj.__class__.__name__}")
            self.console.print(table)

    def print_history(self, entities):
        table = Table(box=self.box)
        print("History:")
        for column in ("date", "type", "old_value", "new_value"):
            table.add_column(column)
        for event in entities:
            table.add_row(event.date.strftime("%Y-%m-%d %H:%M"),
                          event.type.name, event.old_value, event.new_value)
        if table.row_count:
            self.console.print(table)

    def print_commentaries(self, entities):
        table = Table(box=self.box)
        print("Comments:")
        for column in ("date", "text"):
            table.add_column(column)
        entities.sort(key=lambda c: c.date, reverse=False)
        for event in entities:
            table.add_row(event.date.strftime("%Y-%m-%d %H:%M"), event.text)
        if table.row_count:
            self.console.print(table)

    def print_entity(self,
                     task,
                     entity_type,
                     entities,
                     repo,
                     print_options: PrintOptions = PrintOptions(),
                     kwargs: Optional[Dict[str, Any]] = None):
        if entity_type == "sprints":
            if entities:
                self.print_sprint(entities, repo, print_options, kwargs)
            else:
                exit(f"No sprint with id '{task}' found!")
        if entity_type in ("epics", "stories"):
            if entities:
                self.print_composite(entities, repo, print_options,
                                     entity_type, kwargs)
            else:
                exit(f"No {entity_type} with id '{task}' found!")
        if entity_type == "projects":
            if entities:
                self.print_project(entities, print_options, kwargs)
            else:
                exit(f"No project '{task}' found!")
        if entity_type == "tasks":
            if entities:
                self.print_task(entities, repo, print_options)
            else:
                print(f"No task with id '{task}' found!")
        if entity_type == "notes":
            if entities:
                self.print_note(entities=entities,
                                association="task",
                                markdown=True)
            else:
                print(f"No notes with id '{task}' found!")

    def print_entities(self,
                       entities,
                       type,
                       repo,
                       custom_sort,
                       print_options=PrintOptions()):
        if type == "projects":
            entities.sort(key=self._sort_open_tasks, reverse=True)
            self.print_project(entities, print_options)
        elif type == "tasks":
            if custom_sort and custom_sort not in ("status", "priority"):
                entities.sort(key=lambda c: getattr(c, custom_sort),
                              reverse=False)
            elif custom_sort == "status":
                entities.sort(key=lambda c: c.status.value, reverse=True)
            elif custom_sort == "priority":
                entities.sort(key=lambda c: c.priority.value, reverse=True)
            else:
                entities.sort(key=lambda c: (c.status.value, c.priority.value),
                              reverse=True)
            self.print_task(entities=entities,
                            repo=repo,
                            print_options=print_options,
                            custom_sort=custom_sort)
        elif type == "tags":
            self.print_tag(entities=entities)
        elif type == "users":
            self.print_user(entities=entities)
        elif type == "sprints":
            self.print_sprint(entities=entities,
                              repo=repo,
                              print_options=print_options)
        elif type in ("epics", "stories"):
            if custom_sort == "project":
                entities.sort(key=lambda c: c.project, reverse=True)
            elif custom_sort == "tasks":
                entities.sort(key=lambda c: len(c.tasks), reverse=True)

            self.print_composite(entities=entities,
                                 repo=repo,
                                 print_options=print_options,
                                 composite_type=type)
        elif type == "notes":
            self.print_note(entities=entities, association="task")
        else:
            self.print_default_entity(self, entities)

    def print_user(self, entities):
        table = Table(box=self.box)
        for column in ("id", "name"):
            table.add_column(column)
        seen_users = set()
        for entity in entities:
            if entity.name not in seen_users:
                table.add_row(f"[red]{entity.id}[/red]", entity.name)
                seen_users.add(entity.name)
        if table.row_count:
            self.console.print(table)

    def print_tag(self, entities):
        table = Table(box=self.box)
        for column in ("id", "text"):
            table.add_column(column)
        seen_tags = set()
        for entity in entities:
            if entity.text not in seen_tags:
                table.add_row(f"[red]{entity.id}[/red]", entity.text)
                seen_tags.add(entity.text)
        if table.row_count:
            self.console.print(table)

    def print_note(self, entities, association: str, markdown=False) -> None:

        association = next(iter(entities)).__class__.__name__.replace(
            "Note", "").lower()
        table = Table(box=self.box, title="NOTES")
        for column in ("id", "name", "association"):
            table.add_column(column)
        for entity in entities:
            if markdown:
                if isinstance(entity.text, (bytes, bytearray)):
                    text = entity.text.decode("utf-8")
                else:
                    text = entity.text
                self.console.print(Markdown(text), width=80)
                exit()
            else:
                table.add_row(f"[red]{str(entity.id)}[/red]", entity.name,
                              str(getattr(entity, association)))
        if table.row_count:
            self.console.print(table)

    def print_default_entity(self, entities):
        table = Table(box=self.box)
        for column in ("id", "date", "text"):
            table.add_column(column)
        for entity in entities:
            table.add_row(f"[red]{entity.id}[/red]",
                          entity.date.strftime("%Y-%m-%d %H:%M"), entity.text)
        if table.row_count:
            self.console.print(table)

    def print_composite(self,
                        entities,
                        repo,
                        print_options,
                        composite_type,
                        kwargs=None):
        table = Table(box=self.box,
                      title=composite_type.upper(),
                      expand=print_options.expand_table)
        non_active_entities = Table(box=self.box,
                                    title=f"INACTIVE {composite_type.upper()}",
                                    expand=print_options.expand_table)
        default_columns = ("id", "name", "description", "status", "project",
                           "tasks")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = default_columns
        for column in printable_columns:
            table.add_column(column, style="bold")
        for column in printable_columns:
            non_active_entities.add_column(column, style="bold")
        for i, entity in enumerate(entities):
            project = services.lookup_project_name(
                entity.project, repo)
            tasks = []
            completed_tasks = []
            for entity_task in entity.tasks:
                task = entity_task.tasks
                if task.status.name not in ("DONE", "DELETED"):
                    tasks.append(task)
                else:
                    completed_tasks.append(task)
            printable_row = {
                "id": f"{entity.id}",
                "name": str(entity.name),
                "description": entity.description,
                "status": entity.status.name,
                "project": project,
                "tasks": str(len(tasks))
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in printable_columns
            ]
            table.add_row(*printable_elements)
            if entity.status.name == "COMPLETED":
                non_active_entities.add_row(*printable_elements)
                continue
            if len(completed_tasks) == len(entity.tasks):
                non_active_entities.add_row(*printable_elements)
                continue
            # if i == 0 or tasks or print_options.show_completed:
            # TODO: Uncomment
            # table.add_row(*printable_elements)
        if table.row_count:
            self.console.print(table)
        if non_active_entities.row_count and print_options.show_completed:
            self.console.print(non_active_entities)
        if print_options.show_tasks and tasks:
            self.print_task(entities=tasks,
                            repo=repo,
                            print_options=print_options,
                            show_window=False,
                            view_level=composite_type[:-1],
                            kwargs=kwargs)
        if print_options.show_tasks and completed_tasks:
            self.print_task(entities=completed_tasks,
                            repo=repo,
                            print_options=print_options,
                            show_window=False,
                            kwargs=kwargs)
        if i == 0 and print_options.show_commentaries and (
                commentaries := entity.commentaries):
            self.print_commentaries(commentaries)
        if viz := print_options.show_viz:
            if "time" in viz:
                time_entries = entity.daily_time_entries_hours()
                self._print_time_utilization(time_entries)

    def print_sprint(self, entities, repo, print_options, kwargs=None):
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                      expand=print_options.expand_table)
        columns = ("id", "start_date", "end_date", "goal", "status",
                   "open tasks", "pct_completed", "velocity", "collaborators",
                   "time_spent", "utilization")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = columns

        for column in printable_columns:
            table.add_column(column)
        for i, entity in enumerate(entities):
            story_points = []
            tasks = []
            collaborators = []
            collaborators_dict = defaultdict(int)
            for sprint_task in entity.tasks:
                story_points.append(sprint_task.story_points)
                task = sprint_task.tasks
                tasks.append(task)
                if task_collaborators := task.collaborators:
                    for collaborator in task.collaborators:
                        collaborators_dict[collaborator.users.
                                           name] += sprint_task.story_points
                else:
                    collaborators_dict["Unknown"] += sprint_task.story_points

            for user, story_point in sorted(collaborators_dict.items(),
                                            key=lambda x: x[1],
                                            reverse=True):
                collaborators.append(f"{user} ({round(story_point, 2)})")

            collaborators_string = ", ".join(collaborators)
            open_tasks = [
                task for task in tasks
                if task.status.name not in ("DONE", "DELETED")
            ]
            if (total_tasks := len(tasks)) > 0:
                pct_completed = round(
                    (len(tasks) - len(open_tasks)) / len(tasks) * 100)
            else:
                pct_completed = 0
            printable_row = {
                "id": str(entity.id),
                "start_date": str(entity.start_date),
                "end_date": str(entity.end_date),
                "goal": entity.goal,
                "status": entity.status.name,
                "open tasks": f"{len(entity.open_tasks)} ({len(entity.tasks)})",
                "pct_completed": f"{round(entity.pct_completed * 100, 2)}%",
                "velocity": str(round(entity.velocity, 2)),
                "collaborators": collaborators_string,
                "time_spent": str(formatter.Formatter.format_time_spent(entity.total_time_spent)),
                "utilization": f"{round(entity.utilization * 100)}%"
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in printable_columns
            ]
            table.add_row(*printable_elements)
        if table.row_count:
            self.console.print(table)

        if print_options.show_tasks:
            task_print_options = deepcopy(print_options)
            task_print_options.show_commentaries = False
            task_print_options.show_viz = False
            self.print_task(entities=tasks,
                            repo=repo,
                            print_options=task_print_options,
                            story_points=story_points,
                            kwargs=kwargs,
                            show_window=False,
                            view_level="sprint")
        if i == 0 and print_options.show_commentaries and (
                commentaries := entities[0].commentaries):
            self.print_commentaries(commentaries)
        if viz := print_options.show_viz:
            if "cfd" in viz:
                dates = [
                    date.strftime("%Y-%m-%d") for date in pd.date_range(
                        entity.start_date,
                        entity.end_date).to_pydatetime().tolist()
                ]
                status_changes = views.status_changes(repo.session, 2)
                placeholders = pd.DataFrame(data=list(
                    itertools.product([task.id for task in tasks], dates)),
                                            columns=["task", "date"])
                history = self._restore_status_history(placeholders,
                                                       status_changes)
                # TODO: aggreggate history ty date and number of tasks within the state
                # plt.date_form('Y-m-d')
                # y = range(1, 10)
                # y2 = range(2, 20, 2)
                # y3 = range(3, 30, 3)
                # y4 = range(4, 40, 4)
                # y5 = range(5, 50, 5)
                # plt.plot(dates, y2,  fillx=True, color="green", label="TODO")
                # plt.plot(dates, y3,  fillx=False, color="blue", label="IN_PROGRESS")
                # plt.plot(dates, y4,  fillx=False, color="red", label="REVIEW")
                # plt.plot(dates, y5,  fillx=False, color="black", label="DONE")
                # plt.plot_size(50, 15)
                # plt.show()
            if "time" in viz:
                time_entries = entity.daily_time_entries_hours()
                self._print_time_utilization(time_entries)
        if len(entities) == 1:
            app = TerkaSprint(entity=entities[0], repo=self.repo, config=self.config)
            app.run()


    def print_project(self, entities, print_options, kwargs=None):
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                      expand=print_options.expand_table)
        non_active_projects = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                                    expand=print_options.expand_table)
        default_columns = ("id", "name", "description", "status", "open_tasks",
                           "overdue", "stale", "backlog", "todo",
                           "in_progress", "review", "done", "median_task_age",
                           "time_spent")

        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = default_columns
        for column in printable_columns:
            table.add_column(column)
        for column in printable_columns:
            if column in ("id", "name", "description", "status", "open_tasks"):
                non_active_projects.add_column(column)
        for entity in entities:
            # if entity.status.name != "ACTIVE":
            #     continue
            open_tasks = self._sort_open_tasks(entity)
            if open_tasks > 0 and entity.status.name == "ACTIVE":
                overdue_tasks = [
                    task for task in entity.tasks
                    if task.due_date and task.due_date <= datetime.now().date(
                    ) and task.status.name not in ("DELETED", "DONE")
                ]
                stale_tasks = []
                for task in entity.tasks:
                    if task.is_stale:
                        stale_tasks.append(task)
                stale_tasks = list(set(stale_tasks))
                median_task_age = round(
                    median([(datetime.now() - task.creation_date).days
                            for task in entity.tasks
                            if task.status.name not in ("DELETED", "DONE")]))
                backlog = self._count_task_status(entity.tasks, "BACKLOG")
                todo = self._count_task_status(entity.tasks, "TODO")
                in_progress = self._count_task_status(entity.tasks,
                                                      "IN_PROGRESS")
                review = self._count_task_status(entity.tasks, "REVIEW")
                done = self._count_task_status(entity.tasks, "DONE")

                if overdue_tasks:
                    entity_id = f"[red]{entity.id}[/red]"
                else:
                    entity_id = f"[green]{entity.id}[/green]"
                printable_row = {
                    "id":
                    f"{entity.id}",
                    "name":
                    str(entity.name),
                    "description":
                    entity.description,
                    "status":
                    entity.status.name,
                    "open_tasks":
                    str(open_tasks),
                    "overdue":
                    str(len(overdue_tasks)),
                    "stale":
                    str(len(stale_tasks)),
                    "backlog":
                    str(backlog),
                    "todo":
                    str(todo),
                    "in_progress":
                    str(in_progress),
                    "review":
                    str(review),
                    "done":
                    str(done),
                    "median_task_age":
                    str(median_task_age),
                    "time_spent":
                    formatter.Formatter.format_time_spent(entity.total_time_spent),
                }
                printable_elements = [
                    value for key, value in printable_row.items()
                    if key in printable_columns
                ]
                table.add_row(*printable_elements)
            else:
                printable_row = {
                    "id": f"{entity.id}",
                    "name": str(entity.name),
                    "description": entity.description,
                    "status": entity.status.name,
                    "open_tasks": str(open_tasks),
                }
                printable_elements = [
                    value for key, value in printable_row.items()
                    if key in printable_columns
                ]
                non_active_projects.add_row(*printable_elements)

        if table.row_count:
            self.console.print(table)
        if non_active_projects.row_count:
            self.console.print("[green]Inactive projects[/green]")
            self.console.print(non_active_projects)
        non_task_view_options = deepcopy(print_options)
        non_task_view_options.show_tasks = False
        non_task_view_options.show_viz = False
        if print_options.show_epics and (epics := entity.epics):
            self.console.print("")
            self.print_composite(epics, self.repo, non_task_view_options,
                                 "epics", kwargs)
        if print_options.show_stories and (stories := entity.stories):
            self.console.print("")
            self.print_composite(stories, self.repo, non_task_view_options,
                                 "stories", kwargs)
        if print_options.show_tasks and (tasks := entity.tasks):
            task_print_options = deepcopy(print_options)
            task_print_options.show_commentaries = False
            task_print_options.show_viz = False
            self.console.print("")
            self.print_task(entities=tasks,
                            repo=self.repo,
                            print_options=task_print_options,
                            show_window=False,
                            kwargs=kwargs)
        if print_options.show_commentaries and (commentaries :=
                                                entity.commentaries):
            self.print_commentaries(commentaries)
        if print_options.show_history and (history := entity.history):
            self.print_history(history)
        if print_options.show_notes and (notes := entity.notes):
            self.print_note(notes, "project")
        if viz := print_options.show_viz:
            if "time" in viz:
                time_entries = entity.daily_time_entries_hours()
                self._print_time_utilization(time_entries)
        if len(entities) == 1:
            collaborators = defaultdict(int)
            for task in entities[0].tasks:
                if task_collaborators := task.collaborators:
                    for collaborator in task.collaborators:
                        name = collaborator.users.name
                        collaborators[name] = +task.total_time_spent
                else:
                    collaborators["me"] = +task.total_time_spent

            app = TerkaProject(entity=entities[0], repo=self.repo, config=self.config)
            app.run()

    def print_task(self,
                   entities,
                   repo,
                   print_options,
                   custom_sort=None,
                   show_window=True,
                   story_points=None,
                   view_level="tasks",
                   kwargs=None):
        table = Table(box=self.box, title="TASKS")
        if story_points:
            default_columns = ("id", "name", "description", "story_points",
                               "status", "priority", "project", "due_date",
                               "tags", "collaborators", "time_spent")
        else:
            default_columns = ("id", "name", "description", "status",
                               "priority", "project", "due_date", "tags",
                               "collaborators", "time_spent")
        completed_tasks_default_columns = ("id", "name", "description",
                                           "status", "priority", "project",
                                           "completed_date", "tags",
                                           "collaborators", "time_spent")
        #TODO: Add printing for only a single task
        # if (entities[0].status.name == "DONE"):
        #     console.print(f"[green]task is completed on [/green]")
        # else:
        #     active_in_days = datetime.now() - entities[0].creation_date
        #     console.print(f"[blue]task is active {active_in_days.days} days[/blue]")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = default_columns
        printable_completed_tasks_colums = printable_columns
        entities = list(entities)
        if kwargs:
            custom_sort = custom_sort or kwargs.pop("sort", None)
            entities = self._get_filtered_entities(entities, kwargs)
        else:
            custom_sort = None
        # if not story_points:
        #     if custom_sort and custom_sort not in ("status", "priority"):
        #         entities.sort(key=lambda c: getattr(c, custom_sort),
        #                       reverse=False)
        #     elif custom_sort == "status":
        #         entities.sort(key=lambda c: c.status.value, reverse=True)
        #     elif custom_sort == "priority":
        #         entities.sort(key=lambda c: c.priority.value, reverse=True)
        #     else:
        #         entities.sort(key=lambda c: (c.status.value, c.priority.value),
        #                       reverse=True)
        table, completed_tasks, completed_story_points = self._print_task(
            table=table,
            entities=entities,
            default_columns=printable_columns,
            repo=repo,
            story_points=story_points,
            show_window=show_window,
            view_level=view_level,
            custom_sort=custom_sort)
        if table.row_count:
            self.console.print(table)
        if print_options.show_completed and completed_tasks:
            table = Table(box=self.box)
            table, *rest = self._print_task(
                table=table,
                entities=completed_tasks,
                default_columns=printable_completed_tasks_colums,
                repo=repo,
                story_points=completed_story_points,
                show_window=show_window,
                all_tasks=False,
                view_level=view_level,
                custom_sort=custom_sort)
            if table.row_count:
                self.console.print(f"[green]****COMPLETED TASKS*****[/green]")
                self.console.print(table)
        if viz := print_options.show_viz:
            if "time" in viz:
                time_entries = entity.daily_time_entries_hours(last_n_days=14)
                self._print_time_utilization(time_entries)

    def _get_attributes(self, obj) -> List[Tuple[str, str]]:
        import inspect
        attributes = []
        for name, value in inspect.getmembers(obj):
            if not name.startswith("_") and not inspect.ismethod(value):
                if not value:
                    continue
                elif name in ("created_by", "assignee"):
                    attributes.append(
                        (name, services.lookup_user_name(value, self.repo)))
                elif name == "project":
                    project = services.lookup_project_name(value, self.repo)
                    attributes.append((name, project))
                elif hasattr(value, "name"):
                    attributes.append((name, value.name))
                elif isinstance(value, datetime):
                    attributes.append((name, value.strftime("%Y-%m-%d %H:%M")))
                elif isinstance(value, set):
                    values = value.pop()
                    attributes.append((name, str(values)))
                else:
                    attributes.append((name, str(value)))
        return attributes

    def _sort_open_tasks(self, entities):
        return len([
            task for task in entities.tasks
            if task.status.name not in ("DONE", "DELETED")
        ])

    def _count_task_status(self, tasks, status: str) -> int:
        return len([task for task in tasks if task.status.name == status])

    def _get_filtered_entities(self, entities, kwargs):
        if kwargs:
            filtering_attributes = set(list(kwargs))
            temp_entities = []
            attributes = [
                name for name, value in inspect.getmembers(entities[0])
                if not name.startswith("_") and not inspect.ismethod(value)
            ]
            filtering_attributes = filtering_attributes.intersection(
                set(attributes))
            attribute_getter = attrgetter(*filtering_attributes)
            filtered_entities = []
            for entity in entities:
                should_add = False
                for key, value in kwargs.items():
                    if key in filtering_attributes:
                        attribute_value = getattr(entity, key)
                        if hasattr(attribute_value, "name"):
                            if attribute_value.name != value:
                                should_add = False
                                break
                            else:
                                should_add = True
                                continue
                        elif attribute_value != value:
                            should_add = False
                            break
                        else:
                            should_add = True
                            continue
                if should_add:
                    filtered_entities.append(entity)
            entities = list(filtered_entities)
        return entities

    def _print_task(self,
                    table,
                    entities,
                    default_columns,
                    repo,
                    story_points=None,
                    all_tasks=True,
                    show_window=True,
                    view_level="tasks",
                    custom_sort=None):
        if all_tasks:
            completed_tasks = []
        else:
            completed_tasks = None
            completed_story_points = None
        if not custom_sort:
            custom_sort = "status"
            reverse = True
        elif custom_sort == "priority":
            reverse = True
        else:
            reverse = False
        sorting = attrgetter(custom_sort)

        def sorting_fn(x):
            if hasattr(sorting(x), "value"):
                return sorting(x).value
            return sorting(x)

        if story_points:
            completed_story_points = []
            entities = [(entity, story_point) for entity, story_point in
                        sorted(zip(entities, story_points),
                               key=lambda x: sorting_fn(x[0]),
                               reverse=reverse)]
        else:
            entities.sort(key=lambda x: sorting_fn(x), reverse=reverse)
            completed_story_points = None
        for column in default_columns:
            table.add_column(column)
        for entity in entities:
            if story_points:
                story_point = entity[1]
                entity = entity[0]
            else:
                story_point = None
            if tags := entity.tags:
                tag_texts = []
                for tag in list(tags):
                    tag_text = tag.base_tag.text
                    if not tag_text.startswith(view_level):
                        if tag_text.startswith("sprint"):
                            tag_texts.append(f"[yellow]{tag_text}[/yellow]")
                        elif tag_text.startswith("epic"):
                            tag_texts.append(f"[green]{tag_text}[/green]")
                        elif tag_text.startswith("story"):
                            tag_texts.append(f"[magenta]{tag_text}[/magenta]")
                        elif tag_text.startswith("bug"):
                            tag_texts.append(f"[red]{tag_text}[/red]")
                        else:
                            tag_texts.append(tag_text)
                tag_string = ",".join(sorted(tag_texts))
            else:
                tag_string = ""
            if collaborators := entity.collaborators:
                collaborators_texts = sorted([
                    collaborator.users.name
                    for collaborator in list(collaborators)
                    if collaborator.users
                ])
                collaborator_string = ",".join(collaborators_texts)
            else:
                collaborator_string = ""
            if comments := entity.commentaries:
                comments.sort(key=lambda c: c.date, reverse=False)
            if history := entity.history:
                history.sort(key=lambda c: c.date, reverse=False)
            project = services.lookup_project_name(
                entity.project, repo)
            priority = entity.priority.name if hasattr(entity.priority,
                                                       "name") else "UNKNOWN"
            time_spent = formatter.Formatter.format_time_spent(
                entity.total_time_spent)
            if entity.status.name in ("DELETED", "DONE") and all_tasks:
                completed_tasks.append(entity)
                if story_points:
                    completed_story_points.append(story_point)
                continue
            entity_name = f"{entity.name}" if not comments else f"{entity.name} [blue][{len(comments)}][/blue]"

            completed_date = None
            if event_history := entity.history:
                completed_events = [
                    event.date for event in event_history
                    if event.new_value in ("DONE", "DELETED")
                ]
                if completed_events:
                    completed_date = max(completed_events).strftime("%Y-%m-%d")
            if not all_tasks:
                entity_id = str(entity.id)
            elif entity.is_overdue:
                entity_id = f"[red]{entity.id}[/red]"
            elif entity.is_stale:
                entity_id = f"[yellow]{entity.id}[/yellow]"
            else:
                entity_id = str(entity.id)

            printable_row = {
                "id": entity_id,
                "name": entity_name,
                "description": entity.description,
                "story_points": str(story_point),
                "status": entity.status.name,
                "priority": priority,
                "project": project,
                "due_date": str(entity.due_date or completed_date),
                "tags": tag_string,
                "collaborators": collaborator_string,
                "time_spent": str(time_spent)
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in default_columns
            ]
            if story_points:
                table.add_row(*printable_elements)
            else:
                table.add_row(*printable_elements)
        if table.row_count == 1 and show_window:
            app = TerkaTask(entity=entity,
                            project=project,
                            history=history,
                            commentaries=comments)
            app.run()
        # # TODO: not working
        # if print_options.show_commentaries and (commentaries :=
        #                                         entity.commentaries):
        #     self.print_commentaries(commentaries)
        # # TODO: not working
        # if print_options.show_history and (history := entity.history):
        #     self.print_history(history)
        return table, completed_tasks, completed_story_points

    def _restore_status_history(self, placeholders, status_history):
        partial_history = status_history[
            status_history["last_status_for_date"].notnull()]
        current_values = status_history[["task",
                                         "current_status"]].drop_duplicates()
        if not partial_history.empty:
            joined = pd.merge(placeholders,
                              partial_history,
                              on=["task", "date"],
                              how="left")
            joined = joined.replace({np.nan: None})
            joined["filled_backward"] = joined.groupby(
                "task")["first_status_for_date"].bfill()
            joined["filled_forward"] = joined.groupby(
                "task")["last_status_for_date"].ffill()
            joined["filled_forward"] = joined.groupby(
                "task")["filled_forward"].fillna(joined.pop("filled_backward"))
            joined = pd.merge(joined, current_values, on="task", how="left")
            joined["status"] = joined["filled_forward"]
        else:
            joined = pd.merge(placeholders,
                              current_values,
                              on="task",
                              how="left")
        return joined[["date", "task", "status"]]

    def _print_time_utilization(self, time_entries) -> None:
        plt.date_form('Y-m-d')
        plt.plot_size(100, 15)
        plt.title(f"Time tracker - {format_hour_minute(sum(time_entries.values()))} spent")
        plt.bar(time_entries.keys(), time_entries.values())
        plt.show()


def format_hour_minute(time_minutes: int) -> str:
    if time_minutes < 1:
        return f"00H: {round(time_minutes * 60)}M"
    else:
        return f"{round(time_minutes // 1)}H:{round(time_minutes % 1* 60)}M"



from typing import Dict, Literal, Optional, Type
from dataclasses import dataclass, asdict


@dataclass
class Command:

    def get_only_set_attributes(self) -> dict:
        return {key: value for key, value in asdict(self).items() if value}

    @classmethod
    def from_kwargs(cls, **kwargs: dict) -> Type["Command"]:
        return cls(**{
            k: v
            for k, v in kwargs.items() if k in cls.__match_args__ and v
        })


# Base Commands
@dataclass
class Note(Command):
    id: int
    text: str


@dataclass
class Comment(Command):
    id: int
    text: str


@dataclass
class Complete(Command):
    id: int
    comment: str | None = None


@dataclass
class Delete(Command):
    id: int
    comment: str | None = None


@dataclass
class Tag(Command):
    id: int
    tag: int


@dataclass
class Edit(Command):
    ...


@dataclass
class Create(Command):
    ...


@dataclass
class Update(Command):
    ...


@dataclass
class Start(Command):
    ...


@dataclass
class Create(Command):
    ...


@dataclass
class List(Command):
    ...


@dataclass
class Show(Command):
    id: int


@dataclass
class Get(Command):
    ...


@dataclass
class Collaborate(Command):
    id: int
    collaborator: str


@dataclass
class Track(Command):
    ...


@dataclass
class Add(Command):
    ...


@dataclass
class Connect(Command):
    ...


@dataclass
class Report(Command):
    ...


# TASKS
@dataclass
class CreateTask(Command):
    name: str | None = None
    description: str | None = None
    project: str | None = None
    # tags: str | None = None
    # collaborators: str | None = None
    assignee: str | None = None
    due_date: str | None = None
    status: str = "BACKLOG"
    priority: str = "NORMAL"
    # sprint_id: str | None = None
    # epic_id: str | None = None
    # story_id: str | None = None


@dataclass
class CompleteTask(Complete):
    hours: int | None = None


@dataclass
class DeleteTask(Delete):
    hours: int | None = None
    sprint: str | None = None
    epic: str | None = None
    story: str | None = None


@dataclass
class UpdateTask(Command):
    id: int
    name: str | None = None
    description: str | None = None
    project: str | None = None
    # tags: str | None = None
    # collaborators: str | None = None
    assignee: str | None = None
    due_date: str | None = None
    status: str | None = None
    priority: str | None = None

    # sprint_id: str | None = None
    # epic_id: str | None = None
    # story_id: str | None = None

    def __bool__(self) -> bool:
        return all(f for f in self.__dataclass_fields__ if f != "id")


@dataclass
class TagTask(Tag):
    ...


@dataclass
class CollaborateTask(Collaborate):
    ...


@dataclass
class AssignTask(Command):
    id: int
    user: str


@dataclass
class AddTask(Command):
    id: int
    sprint: str | None = None
    epic: str | None = None
    story: str | None = None


@dataclass
class NoteTask(Note):
    ...


@dataclass
class CommentTask(Comment):
    ...


# PROJECT
@dataclass
class CreateProject(Command):
    name: str | None = None
    description: str | None = None
    workspace: int = 1
    status: str = "ACTIVE"


@dataclass
class UpdateProject(Command):
    id: int
    name: str | None = None
    description: str | None = None
    workspace: str | None = None
    # tags: str | None = None
    # collaborators: str | None = None
    status: str | None = None


@dataclass
class CompleteProject(Complete):
    ...


@dataclass
class DeleteProject(Delete):
    ...


@dataclass
class CommentProject(Comment):
    ...


@dataclass
class TagProject(Tag):
    ...


@dataclass
class ShowProject(Show):
    ...


@dataclass
class ListProject(List):
    ...


# SPRINT
@dataclass
class CreateSprint(Command):
    goal: str | None = None
    start_date: str | None = None
    end_date: str | None = None

    def __bool__(self) -> bool:
        if self.start_date and self.end_date:
            return True
        return False


@dataclass
class StartSprint(Command):
    id: int


@dataclass
class CompleteSprint(Complete):
    ...


@dataclass
class ListSprint(List):
    ...


# EPIC
@dataclass
class CreateEpic(Command):
    name: str | None = None
    description: str | None = None
    project: str | None = None


@dataclass
class CompleteEpic(Complete):
    ...


@dataclass
class DeleteEpic(Delete):
    ...


@dataclass
class UpdateEpic(Command):
    id: int
    name: str | None = None
    description: str | None = None
    project: str | None = None


@dataclass
class CommentEpic(Comment):
    ...


@dataclass
class AddEpic(Command):
    id: int
    sprint_id: int


@dataclass
class ListEpic(List):
    ...


# STORIES
@dataclass
class CreateStory(Command):
    name: str | None = None
    description: str | None = None
    project: str | None = None


@dataclass
class CompleteStory(Complete):
    ...


@dataclass
class DeleteStory(Delete):
    ...


@dataclass
class UpdateStory(Command):
    id: int
    name: str | None = None
    description: str | None = None
    project: str | None = None


@dataclass
class CommentStory(Comment):
    ...


@dataclass
class AddStory(Command):
    id: int
    sprint_id: int


# WORKSPACES
@dataclass
class CreateWorkspace(Command):
    name: str | None = None
    description: str | None = None


@dataclass
class DeleteWorkspace(Delete):
    ...


# TAGS
@dataclass
class ShowTag(Show):
    ...


@dataclass
class ListTag(List):
    ...


# USERS
@dataclass
class ShowUser(Show):
    ...


@dataclass
class ListUser(List):
    ...



class Formatter:

    @staticmethod
    def calculate_time_spent(entity) -> str:
        time_spent_sum = sum(
            [entry.time_spent_minutes for entry in entity.time_spent])
        return Formatter.format_time_spent(time_spent_sum)

    @staticmethod
    def format_time_spent(time_spent: int) -> str:
        time_spent_hours = time_spent // 60
        time_spent_minutes = time_spent % 60
        if time_spent_hours and time_spent_minutes:
            time_spent = f"{time_spent_hours}H:{time_spent_minutes}M"
        elif time_spent_hours:
            time_spent = f"{time_spent_hours}H:00M"
        elif time_spent_minutes:
            time_spent = f"00H:{time_spent_minutes}M"
        else:
            time_spent = ""
        return time_spent



from typing import Set
from collections import defaultdict
from enum import Enum
import logging
from datetime import datetime, date
from dataclasses import dataclass

from .task import Task

logger = logging.getLogger(__name__)


class SprintStatus(Enum):
    PLANNED = 1
    ACTIVE = 2
    COMPLETED = 3


class Sprint:

    def __init__(self,
                 start_date: datetime = None,
                 end_date: datetime = None,
                 status: str = "PLANNED",
                 goal: str = None,
                 **kwargs) -> None:
        if not start_date and not end_date:
            raise ValueError("Please add start and end date of the sprint")
        if start_date.date() < datetime.today().date():
            raise ValueError(f"start date cannot be less than today")
        if not start_date:
            raise ValueError(
                "Please provide start date for the sprint in YYYY-MM-DD format"
            )
        self.start_date = start_date
        if not end_date:
            raise ValueError(
                "Please provide end date for the sprint in YYYY-MM-DD format")
        self.end_date = end_date
        if end_date.date() < start_date.date():
            raise ValueError(f"Sprint end date cannot be less than start date")
        if end_date.date() < datetime.today().date():
            raise ValueError(f"Sprint end date cannot be less than today")
        self.status = self._validate_status(status)
        self.goal = goal
        self.is_completed = False

    def _validate_status(self, status):
        if status and status not in [s.name for s in SprintStatus]:
            raise ValueError(f"{status} is invalid status")
        else:
            return status

    def complete(self, tasks) -> None:
        incompleted_tasks = list()
        for task in tasks:
            if task.tasks.status.name != "DONE":
                incompleted_tasks.append(task)
        if incompleted_tasks:
            logging.warning("[Sprint %d]: %d tasks haven't been completed",
                            self.id, len(incompleted_tasks))
        self.is_completed = True

    @property
    def velocity(self) -> float:
        return sum([t.story_points for t in self.tasks])

    @property
    def utilization(self) -> float:
        if not (velocity := self.velocity) or not (total_time_spent :=
                                                   self.total_time_spent):
            return 0
        return total_time_spent / (velocity * 60)

    @property
    def total_time_spent(self):
        total_time_spent_sprint = 0
        for sprint_task in self.tasks:
            total_time_spent_sprint += sprint_task.tasks.total_time_spent
        return total_time_spent_sprint

    @property
    def open_tasks(self):
        return [
            task for task in self.tasks
            if task.tasks.status.name not in ("DONE", "DELETED")
        ]

    @property
    def pct_completed(self) -> float:
        if (total_tasks := len(self.tasks)) > 0:
            return (total_tasks - len(self.open_tasks)) / total_tasks
        return 0

    @property
    def collaborators(self):
        collaborators = defaultdict(int)
        for sprint_task in self.tasks:
            if task_collaborators := sprint_task.tasks.collaborators:
                for collaborator in task_collaborators:
                    name = collaborator.users.name or "me"
                    collaborators[name] += sprint_task.tasks.total_time_spent
            else:
                collaborators["me"] += sprint_task.tasks.total_time_spent
        return collaborators

    def daily_time_entries_hours(self) -> dict[str, float]:
        entries: dict[str, float] = defaultdict(float)
        for task in self.tasks:
            task_entries = task.tasks.daily_time_entries_hours(
                self.start_date, self.end_date)
            for day, hours in task_entries.items():
                entries[day] += hours
        return entries


@dataclass
class SprintTask:
    sprint: int
    task: int
    story_points: int = 0
    #TODO: add actual_time_spent: int = 0
    is_active_link: bool = True



from typing import Any, Dict, List, Optional, Tuple

from datetime import datetime, timedelta
import re
from terka.service_layer import services
from terka.adapters.repository import AbsRepository


def create_task_dict(kwargs: List[str]) -> Dict[str, str]:
    new_dict = {}
    for i, kwarg in enumerate(kwargs):
        if i == 0 and not kwarg.startswith("-"):
            new_dict["id"] = kwarg
        arr = kwarg.split("=")
        if len(arr) == 2:
            key = re.sub("^-+", "", arr[0])
            new_dict[key] = arr[1]
        else:
            key = arr[0]
            if i < len(kwargs) and key.startswith("-"):
                key = re.sub("^-+", "", key)
                try:
                    new_dict[key] = kwargs[i + 1]
                except IndexError:
                    new_dict[key] = True
    return new_dict


def format_task_dict(config, entity, kwargs) -> Dict[str, Optional[str]]:
    if len(kwargs) > 1:
        new_dict = create_task_dict(kwargs)
        task_dict = {
            "id":
            new_dict.get("id") if entity != "commentaries" else None,
            "task_id":
            new_dict.get("task_id") or new_dict.get("task-id"),
            "name":
            new_dict.get("n") or new_dict.get("name"),
            "num_log_entries":
            new_dict.get("L") or new_dict.get("lines"),
            "status":
            new_dict.get("s") or new_dict.get("status"),
            "project":
            new_dict.get("p") or new_dict.get("project")
            or new_dict.get("project_name"),
            "workspace":
            new_dict.get("w") or new_dict.get("workspace")
            or new_dict.get("workspace_name") or config.get("workspace"),
            "assignee":
            new_dict.get("a") or new_dict.get("assignee"),
            "due_date":
            new_dict.get("d") or new_dict.get("due-date"),
            "description":
            new_dict.get("desc") or new_dict.get("description"),
            "text":
            new_dict.get("t") or new_dict.get("text"),
            "tag":
            new_dict.get("tag") or new_dict.get("tags"),
            "collaborator":
            new_dict.get("collaborator") or new_dict.get("collaborators"),
            "all":
            new_dict.get("all") or new_dict.get("show-completed"),
            "priority":
            new_dict.get("priority"),
            "overdue":
            new_dict.get("overdue"),
            "stale":
            new_dict.get("stale"),
            "goal":
            new_dict.get("goal"),
            "start_date":
            convert_date(new_dict.get("start-date")),
            "end_date":
            convert_date(new_dict.get("end-date")),
            "sprint":
            new_dict.get("to-sprint") or new_dict.get("from-sprint")
            or new_dict.get("sprint") or new_dict.get("sprints"),
            "story":
            new_dict.get("to-story") or new_dict.get("from-story")
            or new_dict.get("story") or new_dict.get("stories"),
            "epic":
            new_dict.get("to-epic") or new_dict.get("from-epic")
            or new_dict.get("epic") or new_dict.get("epics"),
            "story_points":
            new_dict.get("story-points"),
            "hours":
            new_dict.get("H"),
            "minutes":
            new_dict.get("M"),
            "show_history":
            new_dict.get("show-history"),
            "show_commentaries":
            new_dict.get("show-commentaries") or new_dict.get("show-comments"),
            "show_notes":
            new_dict.get("show-notes"),
            # "epics":
            # new_dict.get("epics"),
            # "stories":
            # new_dict.get("stories"),
            # "tasks":
            # new_dict.get("tasks"),
            "external_project":
            new_dict.get("external-project"),
            "external_task":
            new_dict.get("external-task"),
            "show_viz":
            new_dict.get("show-viz"),
            "file":
            new_dict.get("file") or new_dict.get("f"),
            "columns":
            new_dict.get("columns") or new_dict.get("c"),
            "expand":
            new_dict.get("expand"),
            "no-expand":
            new_dict.get("no-expand"),
            "comment":
            new_dict.get("comment"),
        }
        if "--sort" in kwargs:
            sort_index = kwargs.index("--sort")
            sort_statement = kwargs[sort_index:sort_index + 2]
            task_dict.update(create_task_dict(sort_statement))
    elif len(kwargs) == 1:
        if "overdue" in kwargs[0]:
            task_dict = {"overdue": "overdue"}
        elif "stale" in kwargs[0]:
            task_dict = {"stale": 7}
        elif "all" in kwargs[0]:
            task_dict = {"all": "all"}
        elif "tasks" in kwargs[0]:
            task_dict = {"tasks": True}
        else:
            task_dict = {"id": kwargs[0]}
    else:
        task_dict = {}

    if task_dict and isinstance(task_dict, dict):
        task_dict = {k: v for k, v in task_dict.items() if v}

    if "status" in task_dict:
        task_dict["status"] = convert_status(task_dict["status"])
    if "due_date" in task_dict:
        task_dict["due_date"] = convert_date(task_dict["due_date"])
    if not task_dict.get("id"):
        if (id := config.get("task_id")):
            if entity != "commentaries":
                task_dict["id"] = str(config.get("task_id"))
            if entity == "commentaries":
                task_dict["task_id"] = str(config.get("task_id"))

    if not task_dict.get("project") and config.get("project_name"):
        task_dict["project"] = config.get("project_name")
    if any(key in task_dict.keys() for key in ("epics", "stories", "tasks")):
        task_dict["partial_project_view"] = True
    # add_entity_info(task_dict)
    task_dict["expand_table"] = not task_dict.get("no-expand", False)
    task_dict["show_completed"] = task_dict.get("all", False)
    return task_dict


def add_entity_info(task_dict):
    if sprints := task_dict.get("sprints"):
        task_dict["entity_type"] = "sprint"
        task_dict["entity_id"] = sprints
    if epics := task_dict.get("epics"):
        task_dict["entity_type"] = "epic"
        task_dict["entity_id"] = epics
    if stories := task_dict.get("stories"):
        task_dict["entity_type"] = "story"
        task_dict["entity_id"] = stories


def convert_status(status: str):
    conversion_dict = {
        "b": "BACKLOG",
        "t": "TODO",
        "i": "IN_PROGRESS",
        "r": "REVIEW",
        "d": "DONE",
        "x": "DELETED",
        "a": "ACTIVE",
        "p": "PLANNED"
    }
    statuses = [
        conversion_dict.get(status[0].lower(), "BACKLOG")
        for status in status.split(",")
    ]

    return ",".join(statuses)


def convert_date(date: str):
    if not date:
        return date
    if date.startswith("+"):
        due_date = datetime.now().date() + timedelta(days=float(date[1:]))
        return due_date.strftime("%Y-%m-%d")
    if date.startswith("-"):
        due_date = datetime.now().date() - timedelta(days=float(date[1:]))
        return due_date.strftime("%Y-%m-%d")
    return date


def process_command(command: str, config: Dict[str, Any],
                    repo: AbsRepository) -> Tuple[str, str, Dict[Any, Any]]:
    task_dict = {}
    task_list = []
    entity = None
    command, *rest = command.split(" ", maxsplit=2)
    if command[0] == "q":
        exit()
    if rest:
        entity, *task_dict = rest
        if task_dict:
            task_list = re.split(r'\s+(?=[^"]*(?:"[^"]*"[^"]*)*$)',
                                 task_dict[0])
            task_list = [
                re.sub("['|\"]", "", element) for element in task_list
            ]
        task_dict = format_task_dict(config, entity, task_list)
    else:
        task_dict = {}
    task_dict = update_task_dict(task_dict, repo)
    return command, entity, task_dict


def update_task_dict(task_dict: Dict[str, str],
                     repo: AbsRepository) -> Dict[str, str]:
    if (project := task_dict.get("project")):
        project_id = services.lookup_project_id(project, repo)
        task_dict["project"] = project_id
    if (workspace := task_dict.get("workspace")):
        workspace_id = services.lookup_workspace_id(workspace, repo)
        task_dict["workspace"] = workspace_id
    if (assignee := task_dict.get("assignee")):
        task_dict["assignee"] = services.lookup_user_id(assignee, repo)
    task_dict = convert_date_in_task_dict(task_dict)
    return task_dict


def convert_date_in_task_dict(task_dict):
    for key, value in task_dict.items():
        if key.endswith("date"):
            if value == "None":
                task_dict[key] = None
            elif value == "today":
                task_dict[key] = datetime.today()
            else:
                task_dict[key] = datetime.strptime(
                    value, "%Y-%m-%d") if isinstance(value, str) else value
    return task_dict


def format_command(command: str) -> str:
    if command == "d":
        command = input("specify: de[lete], do[ne]")
    if "list".startswith(command):
        return "list"
    if "show".startswith(command):
        return "show"
    if "create".startswith(command):
        return "create"
    if "update".startswith(command):
        return "update"
    if "delete".startswith(command):
        return "delete"
    if "edit".startswith(command):
        return "edit"
    if "comment".startswith(command):
        return "comment"
    return command


def format_entity(entity: str) -> str:
    if entity in "tasks":
        return "task"
    if entity in "projects":
        return "project"
    if entity in "sprints":
        return "sprint"
    if entity in "notes":
        return "note"
    if entity in "epics":
        return "epic"
    if entity == "story" or entity in "stories":
        return "story"
    if entity in "users":
        return "user"
    if entity in "tags":
        return "tag"
    if entity in "workspaces":
        return "workspace"
    return entity



class CollaboratorMixin:
    def __repr__(self):
        return str(self.users.name)


class TaskCollaborator(CollaboratorMixin):

    def __init__(self,
                 id: int,
                 collaborator_id: int,
                 **kwargs: str) -> None:
        self.task = id
        self.collaborator = collaborator_id


class ProjectCollaborator(CollaboratorMixin):

    def __init__(self,
                 id: int,
                 collaborator_id: int,
                 **kwargs: str) -> None:
        self.project = id
        self.collaborator = collaborator_id




from datetime import datetime


class BaseNote:

    def __init__(self,
                 name: str,
                 text: str,
                 created_by: int,
                 date: datetime = datetime.now()) -> None:
        self.name = name
        self.text = text
        self.date = date
        self.created_by = created_by

    def __repr__(self):
        return f"<Note> {self.text}"


class TaskNote(BaseNote):

    def __init__(self,
                 id: int,
                 name: str,
                 text: str,
                 created_by: int,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.task = id
        super().__init__(name=name, text=text, created_by=created_by, date=date)


class ProjectNote(BaseNote):

    def __init__(self,
                 id: int,
                 name: str,
                 text: str,
                 created_by: int,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.project = id
        super().__init__(name=name, text=text, created_by=created_by, date=date)


class EpicNote(BaseNote):

    def __init__(self,
                 id: int,
                 name: str,
                 text: str,
                 created_by: int,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.epic = id
        super().__init__(name=name, text=text, created_by=created_by, date=date)


class StoryNote(BaseNote):

    def __init__(self,
                 id: int,
                 name: str,
                 text: str,
                 created_by: int,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.story = id
        super().__init__(name=name, text=text, created_by=created_by, date=date)


class SprintNote(BaseNote):

    def __init__(self,
                 id: int,
                 name: str,
                 text: str,
                 created_by: int,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.sprint = id
        super().__init__(name=name, text=text, created_by=created_by, date=date)



from typing import Type
from datetime import datetime, timedelta
import re
from subprocess import run
import tempfile

from terka.domain import models, _commands
from terka.utils import convert_date, convert_status


def new_task_template() -> str:
    return f"""
        # You are creating a task, enter below:
        ---
        status: BACKLOG
        name: 
        description: 
        project: 
        due_date: 
        priority: NORMAL
        sprint: 
        epic: 
        story: 
        tags: 
        collaborators: 
        """


def new_composite_template(entity) -> str:
    return f"""
        # You are creating {entity.__name__}, enter below:
        ---
        name: 
        description: 
        project: 
        """


def new_project_template() -> str:
    return f"""
        # You are creating new project, enter below:
        ---
        name: 
        description: 
        workspace: 
        """


def new_sprint_template() -> str:
    today = datetime.now()
    next_monday = (today +
                   timedelta(days=(7 - today.weekday()))).strftime("%Y-%m-%d")
    next_sunday = (today +
                   timedelta(days=(13 - today.weekday()))).strftime("%Y-%m-%d")
    return f"""
        # You are creating a sprint, enter below:
        ---
        goal: 
        description: 
        start_date: {next_monday}
        end_date:  {next_sunday}
        """


def start_sprint_template(sprint: models.sprint.Sprint) -> str:
    ...


def edit_sprint_template(sprint: models.sprint.Sprint) -> str:
    return f"""
        # You are editing sprint {sprint.id}, enter below:
        ---
        goal: {sprint.goal}
        start_date: {sprint.start_date}
        end_date: {sprint.end_date}
        """


def edited_task_template(task: models.task.Task,
                         project: str | None = None) -> str:
    return f"""
        # You are editing task {task.id}, enter below:
        ---
        status: {task.status.name}
        name: {task.name}
        description: {task.description if task.description else ""}
        project: {project or ""}
        sprints: {task.sprints[-1].sprint if task.sprints else ""}
        epics: {task.epics[-1].epic if task.epics else ""}
        stories: {task.stories[-1].story if task.stories else ""}
        tags: {task.tags.pop() if task.tags else ""}
        collaborators: {task.collaborators.pop() if task.collaborators else ""}
        time_spent: 0 (task total_time_spent {task.total_time_spent})
        comment: 
        """


def edited_project_template(task: models.project.Project,
                            workspace: int) -> str:
    return f"""
        # You are editing project {task.id}, enter below:
        ---
        status: {task.status.name}
        name: {task.name}
        description: {task.description if task.description else ""}
        workspace: {workspace}
        tags: {task.tags.pop() if task.tags else ""}
        collaborators: {task.collaborators.pop() if task.collaborators else ""}
        comment: 
        """


def completed_task_template(task: models.task.Task,
                            project: str | None = None) -> str:
    return f"""
        # You are closing task {task.id}, enter below:
        ---
        status: DONE 
        time_spent: 0 (task total_time_spent {task.total_time_spent})
        comment: 
        name: {task.name}
        project: {project or ""}
        description: {task.description if task.description else ""}
        sprints: {task.sprints[-1].sprint if task.sprints else ""}
        epics: {task.epics[-1].epic if task.epics else ""}
        stories: {task.stories[-1].story if task.stories else ""}
        tags: {task.tags.pop() if task.tags else ""}
        collaborators: {task.collaborators.pop() if task.collaborators else ""}
        """


def generate_message_template(entity,
                              kwargs: dict[str, str] | None = None) -> str:
    if not isinstance(entity, type):
        if isinstance(
                entity,
            (models.task.Task, models.story.Story, models.epic.Epic)):
            project = entity.project
        elif isinstance(entity, models.project.Project):
            project = entity.name
            workspace = entity.workspace
        if isinstance(entity, models.sprint.Sprint):
            if not kwargs:
                message_template = new_sprint_template()
            else:
                message_template = edit_sprint_template(entity)
        elif isinstance(entity, models.project.Project):
            message_template = edited_project_template(entity, workspace)
        elif isinstance(entity, models.task.Task):
            message_template = edited_task_template(entity, project)
    else:
        if not kwargs:
            if entity == models.task.Task:
                message_template = new_task_template()
            if entity == models.project.Project:
                message_template = new_project_template()
            if entity == models.sprint.Sprint:
                message_template = new_sprint_template()
            if entity in (models.epic.Epic, models.story.Story):
                message_template = new_composite_template(entity)
        elif kwargs and kwargs.get("status"):
            message_template = completed_task_template(entity, project)
        else:
            message_template = edited_task_template(entity, project)
    return re.sub("\n +", "\n", message_template.lstrip())


def flush_message(entity):
    message_template = generate_message_template(entity)
    with tempfile.NamedTemporaryFile(suffix=".tmp") as tf:
        tf.write(message_template.encode("utf-8"))
        tf.flush()
        run(["vim", "+2", tf.name])
        tf.seek(0)
        new_entry = tf.read()
    return new_entry


def create_command_from_editor(
        entity, command) -> tuple[Type[_commands.Command], dict]:
    new_entry = flush_message(entity)
    new_entry = new_entry.decode("utf-8").rstrip()
    command_arguments: dict = {}
    if entity_id := entity.id:
        command_arguments.update({"id": entity_id})
    for i, row in enumerate(new_entry.split("\n")):
        if not row.startswith("#") and row:
            if row.startswith("--"):
                continue
            entry_type, *entry_value = row.split(": ", maxsplit=2)
            entry_value = ": ".join(entry_value)
            entry_value = entry_value.strip()
            if entry_type == "epics":
                command_arguments["epic_id"] = entry_value
            elif entry_type == "sprints":
                command_arguments["sprint_id"] = entry_value
            elif entry_type == "stories":
                command_arguments["story_id"] = entry_value
            if entry_type == "time_spent":
                time_spent = entry_value.split(" (")
                try:
                    time_spent = time_spent[0].strip()
                    time_spent_minutes = int(time_spent)
                    if time_spent_minutes:
                        command_arguments["hours"] = time_spent_minutes
                except Exception:
                    pass
            elif entry_type == "status":
                command_arguments["status"] = convert_status(entry_value)
            elif entry_type.endswith("date") and entry_value:
                entry_value = datetime.strptime(entry_value, "%Y-%m-%d")
                command_arguments[entry_type] = entry_value
            else:
                command_arguments[entry_type] = entry_value
    return command.from_kwargs(**command_arguments), command_arguments



from __future__ import annotations
from typing import Type

from terka.domain import _commands, events
from terka.service_layer import unit_of_work, handlers

Message = _commands.Command | events.Event


class MessageBus:

    def __init__(self, uow: unit_of_work.AbstractUnitOfWork, publisher,
                 event_handlers: Type[handlers.Handler],
                 command_handlers: Type[handlers.Handler]) -> None:
        self.handler = handlers.Handler(uow, publisher)
        self.event_handlers = event_handlers
        self.command_handlers = command_handlers
        self.return_value = None

    def handle(self, message: Message, context: dict = {}):
        self.queue = [message]
        while self.queue:
            message = self.queue.pop(0)
            if isinstance(message, events.Event):
                self.handle_event(message, context)
            elif isinstance(message, _commands.Command):
                self.handle_command(message, context)

        if self.return_value:
            return self.return_value

    def handle_command(self, command: _commands.Command, context: dict) -> None:
        handler = self.command_handlers[type(command)]
        result = handler(command, self.handler, context)
        self.return_value = result
        self.queue.extend(self.handler.uow.collect_new_events())

    def handle_event(self, event: events.Event, context: dict) -> None:
        for handler in self.event_handlers[type(event)]:
            handler(event, self.handler, context)
            self.queue.extend(self.handler.uow.collect_new_events())



from enum import Enum
from dataclasses import dataclass
from collections import defaultdict
from datetime import date, datetime, timedelta
import pandas as pd


class TaskStatus(Enum):
    DELETED = 0
    BACKLOG = 1
    TODO = 2
    IN_PROGRESS = 3
    REVIEW = 4
    DONE = 5


class TaskPriority(Enum):
    UNKNOWN = 0
    LOW = 1
    NORMAL = 2
    HIGH = 3
    URGENT = 4


@dataclass
class TaskInfo:
    name: str
    description: str
    status: str
    priority: str
    due_date: datetime

class Task:

    _next_task_id = 1

    def __init__(self,
                 name: str,
                 description: str | None = None,
                 creation_date: datetime = datetime.now(),
                 project: int | None = None,
                 assignee: int | None = None,
                 created_by: int | None = None,
                 due_date: datetime | None = None,
                 status: str = "BACKLOG",
                 priority: str = "NORMAL",
                 **kwargs) -> None:
        if not name:
            raise ValueError("task name cannot be empty!")
        if not isinstance(creation_date, datetime):
            raise ValueError(
                "creation_date should be of type datetime.datetime!")
        self._task_id = Task._next_task_id
        Task._next_task_id += 1

        self.name = name
        self.creation_date = creation_date
        self.description = description
        self.project = project
        self.assignee = assignee
        self.created_by = created_by
        if due_date and due_date.date() < datetime.today().date():
            raise ValueError(
                f"Due date {due_date.date()} cannot be less than today")
        else:
            self.due_date = due_date
        self.status = self._validate_status(status)
        if priority and priority not in [p.name for p in TaskPriority]:
            raise ValueError(f"{priority} is invalid priority")
        else:
            self.priority = priority

    def _validate_status(self, status):
        if status and status not in [
                s.name for s in TaskStatus if s.name != "DELETED"
        ]:
            raise ValueError(f"{status} is invalid status")
        else:
            return status

    @property
    def total_time_spent(self):
        if self.time_spent:
            return sum([t.time_spent_minutes for t in self.time_spent])
        return 0

    def daily_time_entries_hours(
            self,
            start_date: str | date | None = None,
            end_date: str | date | None = None,
            last_n_days: int | None = None) -> dict[str, float]:
        if last_n_days:
            start_date = (datetime.today() - timedelta(last_n_days)).date()
            end_date = datetime.today().date()
        elif isinstance(start_date, str) and isinstance(end_date, str):
            start_date = datetime.strptime(start_date, "%Y-%m-%d").date()
            end_date = datetime.strptime(end_date, "%Y-%m-%d").date()
        dates = [
            date.strftime("%Y-%m-%d") for date in pd.date_range(
                start_date, end_date).to_pydatetime().tolist()
        ]
        entries = dict.fromkeys(dates, 0.0)
        for entry in self.time_spent:
            creation_date = entry.creation_date.date()
            if creation_date >= start_date and creation_date <= end_date:
                day = creation_date.strftime("%Y-%m-%d")
                entries[day] += entry.time_spent_minutes / 60
        return entries

    @property
    def completion_date(self) -> datetime | None:
        for event in self.history:
            if event.new_value in ("DONE", "DELETED"):
                return event.date

    @property
    def is_stale(self):
        if self.history and self.status.name in ("TODO", "IN_PROGRESS",
                                                 "REVIEW"):
            if max([event.date for event in self.history
                    ]) < (datetime.today() - timedelta(days=5)):
                return True
        return False

    @property
    def is_overdue(self):
        if self.due_date and self.due_date <= date.today():
            return True
        return False

    def __repr__(self):
        return f"<Task {self.id}>: {self.name}, {self.status.name}, {self.creation_date}"



from typing import Any, Dict, Optional
import logging
import rich
from rich.console import Console
from rich.table import Table
import os
import re
import yaml

from terka.adapters.repository import AbsRepository
from terka.domain import models
# import terka.domain._commands as commands


def update_config(update_pair: Dict[str, Any]):
    config = get_config()
    config.update(update_pair)
    with open(f"{home_dir}/.terka/config.yaml", "w") as f:
        yaml.dump(config, f)


def get_config():
    home_dir = os.path.expanduser('~')
    with open(f"{home_dir}/.terka/config.yaml", "r", encoding="utf-8") as f:
        config = yaml.safe_load(f)
    return config


def lookup_project_id(project_name: str, repo: AbsRepository) -> int:
    projects = project_name.split(",")
    if project_name.startswith("NOT"):
        projects = [
            f"NOT:{project}"
            for project in project_name.replace("NOT:", "").split(",")
        ]
        negate = True
    else:
        projects = project_name.split(",")
        negate = False
    returned_projects = []
    returned_projects_set = set()
    for project_name in projects:
        if not (project := repo.list(models.project.Project, project_name)):
            print(
                f"Creating new project: {project_name}. Do you want to continue (Y/n)?"
            )
            answer = input()
            while answer.lower() != "y":
                print("Provide a project name: ")
                project_name = input()
                print(
                    f"Creating new project: {project_name}. Do you want to continue (Y/n)?"
                )
                answer = input()
            project = models.project.Project(project_name)
            repo.add(project)
            repo.session.commit()
        if isinstance(project, list):
            project_ids = [project.id for project in project]
            if negate:
                if returned_projects_set:
                    returned_projects_set = returned_projects_set.intersection(
                        set(project_ids))
                else:
                    returned_projects_set = returned_projects_set.union(
                        set(project_ids))

            else:
                returned_projects.extend(project_ids)
        else:
            returned_projects.append(project.id)
    if returned_projects_set:
        return list(returned_projects_set)
    if len(returned_projects) == 1:
        return returned_projects[0]
    return returned_projects


def lookup_user_id(user_name: str, repo: AbsRepository) -> int | None:
    if user := repo.get(models.user.User, user_name):
        return user.id
    return None


def lookup_user_name(user_id: str, repo: AbsRepository) -> str | None:
    if user := repo.get_by_id(models.user.User, user_id):
        return user.name
    return None


def lookup_project_name(project_id: int, repo: AbsRepository) -> str | None:
    if project := repo.get_by_id(models.project.Project, project_id):
        return project.name
    return None


def lookup_workspace_id(workspace_name: str,
                        repo: AbsRepository) -> int | None:
    if workspace := repo.get(models.workspace.Workspace, workspace_name):
        return workspace.id
    return None


def get_workplace_by_name(
        workspace_name: str,
        repo: AbsRepository) -> models.workspace.Workspace | None:
    return repo.get(models.workspace.Workspace, workspace_name)


def get_workplace_by_id(
        workspace_id: int,
        repo: AbsRepository) -> models.workspace.Workspace | None:
    return repo.get_by_id(models.workspace.Workspace, workspace_id)


def get_project_by_name(project_name: str,
                        repo: AbsRepository) -> models.project.Project | None:
    return repo.get(models.project.Project, project_name)


def get_project_by_id(project_id: int,
                      repo: AbsRepository) -> models.project.Project | None:
    return repo.get_by_id(models.project.Project, project_id)


# class CommandHander:

#     def __init__(self, home_dir, config, console=Console()):
#         self.home_dir = home_dir
#         self.config = config
#         self.console =  console

#     def execute(self, command: commands.Command, entity, task_dict):
#         # TODO: create registry of commands
#         handler = commands_registry[command]
#         handler.handle(entity, task_dict)


class ServiceCommandHander:

    def __init__(self, home_dir, config, console=Console()):
        self.home_dir = home_dir
        self.config = config
        self.console = console

    def execute(self, command, entity, task_dict):
        is_service_command = True
        if command == "init":
            self.init_terka()
        elif command == "focus":
            self.focus(entity, task_dict)
        elif command == "unfocus":
            self.unfocus()
        elif command == "switch" and entity == "workspace":
            self.switch(task_dict.get("id"))
        elif command == "set" and entity == "workspace":
            self.set(task_dict.get("id"))
        elif command == "log":
            self.log(task_dict)
        else:
            is_service_command = False
        if is_service_command:
            exit()

    def init_terka(self,
                   terka_folder: str = ".terka",
                   default_user: str = "admin",
                   default_config: str = "config.yaml") -> None:
        path = os.path.join(self.home_dir, terka_folder)
        if not os.path.exists(path):
            answer = input(
                f"Do you want to init terka in this directory {path}? [Y/n]")
            if "y" in answer.lower():
                Path(path).mkdir(parents=True, exist_ok=True)
                with open(os.path.join(path, "config.yaml"), "w") as f:
                    yaml.dump({"user": default_user}, f)
            elif "n" in answer.lower():
                path = input("Specify full path to the terka directory: ")
                os.mkdirs(path)
            else:
                exit()
        elif not os.path.exists(os.path.join(path, default_config)):
            answer = input(
                f"Config.yaml not found in {path}, Create it? [Y/n]")
            if "y" in answer.lower():
                with open(os.path.join(path, default_config), "w") as f:
                    yaml.dump({"user": default_user}, f)
            else:
                exit()
        else:
            print("Terka directory already exist.")

    def focus(self, entity_type, kwargs):
        if entity_type == "tasks":
            self.config["task_id"] = kwargs["id"]
            if "project_name" in self.config.keys():
                del self.config["project_name"]
        if entity_type == "projects":
            self.config["project_name"] = kwargs["id"]
            if "task_id" in self.config.keys():
                del self.config["task_id"]
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        logging.info("<focus> %s: %s", entity_type, "")

    def unfocus(self):
        if "task_id" in self.config.keys():
            del self.config["task_id"]
        if "project_name" in self.config.keys():
            del self.config["project_name"]
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        loging.info("<unfocus> %s: %s", "", "")

    def log(self, kwargs):
        table = Table(box=rich.box.SIMPLE)
        with open(f"{self.home_dir}/.terka/terka.log", "r") as f:
            head = f.readlines()
        for column in ("date", "source", "level", "message"):
            table.add_column(column)
        num_log_entries = int(kwargs.get("num_log_entries", 10))
        tail = head[-num_log_entries:]
        for row in tail[::-1]:
            info, message = row.split("] ")
            date, source, level = info.split("][")
            table.add_row(re.sub("\[", "", date), source, level, message)
        self.console.print(table)

    def switch(self, workspace: str):
        if not self.config.get("workspace"):
            self.config["workspace"] = "Default"  # Default workspace
        else:
            self.config["workspace"] = workspace
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        logging.info("switched to workspace: %d", workspace)

    def set_workspace(self, workspace: str):
        self.config["workspace"] = workspace
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        logging.info("set workspace: %d", workspace)
